"""
    An implementation of the Knuth-Bendix algorithm,
    as described in (1), p. 143.
    For determining if two paths in a category are equal.

The algorithm as given here,
takes a set of equations in the form of a sequence:

E = [(a, b), (c, d) ...]

where a, b, c, d are 'paths'.

Paths are given as strings, for example:

E = [ ('fhk', 'gh'), ('m', 'kkm') ]

means that the path 'fhk' equals 'gh' and 'm' equals 'kkm'.

Each arrow in the path is here a single character.  If longer arrow
names are required, a delimiter string can be specified as in:

kb(E, delim='.')

The paths must then be given by the delimiter between each arrow;

E = [ ('h_arrow.g_arrow', 'g_arrow.k_arrow') ... ]


The function kb(E) returns an object, say A, which is

o callable: A(a, b)->boolean determines if two paths given by a, b are equal.
o has a method A.reduce(a)->pathstring, which reduces a path to normal form.

An optional parameter to kb, max_iterations, determines the maximum
number of iterations the algorithm should try making the reduction
system 'confluent'. The algorithm is not guaranteed to terminate
with a confluent system in a finite number of iterations, so if the
number of iterations needed exceeds max_iterations an exception
(ValueError) will be raised. The default is 100.

References

(1)
@book{walters91categories,
    title={Categories and Computer Science},
    author={R. F. C. Walters},
    publisher={Cambridge University Press},
    location={Cambridge},
    year=1991}


(2)
@book{grimaldi94discrete,
author="Ralph P. Grimaldi".
title="Discrete and Combinatorial Mathematics: An Applied Introduction",
publisher="Addison-Wesley",
location="Readin, Massachusetts",
year=1994
}


"""

import functools


class KnuthBendix:
    def __init__(self, E, delim='', max_iterations=100):
        self.reductions = []
        self.delim = delim
        for a, b in E:
            if delim:
                a = self.wrap_delim(a)
                b = self.wrap_delim(b)
            if self.gt(b, a):
                a, b = b, a
            self.reductions.append((a, b))
        self.make_confluent(max_iterations)
        self.sort()

    def __call__(self, x, y):
        return self.reduce(x) == self.reduce(y)

    def gt(self, a, b):
        delim = self.delim
        if delim:
            la = len(a)
            lb = len(b)
        else:
            la = a.count(delim)
            lb = b.count(delim)
        if la > lb:
            return 1
        if la < lb:
            return 0
        return a > b

    def make_confluent(self, max_iterations):
        def add_reduction(p, q):
            if p != q:
                if self.gt(p, q):
                    self.reductions.append((p, q))
                else:
                    self.reductions.append((q, p))
                self.confluent = 0

        reds_tested = {}
        for i in range(max_iterations):
            self.confluent = 1
            reds = list(self.reductions)
            for u1, v1 in reds:
                for u2, v2 in reds:
                    red = (u1, u2, u2, v2)
                    if red in reds_tested:
                        continue
                    reds_tested[red] = 1
                    if u2 in u1:
                        p = self.freduce(v1)
                        i = u1.index(u2)
                        while i >= 0:
                            uuu = u1[:i]+v2+u1[i+len(u2):]
                            q = self.freduce(uuu)
                            add_reduction(p, q)
                            i = u1.find(u2, i+1)
                    lu1 = len(u1)
                    for i in range(1, lu1-len(self.delim)):
                        if u2[:lu1-i] == u1[i:]:
                            p = self.freduce(v1 + u2[lu1-i:])
                            q = self.freduce(u1[:i] + v2)
                            add_reduction(p, q)

            assert ('', '') not in reds
            # Remove redundant reductions
            newr = []
            nullred = (self.delim, self.delim)
            for i, uv in enumerate(self.reductions):
                u, v = uv
                self.reductions[i] = nullred
                ru = self.freduce(u)
                rv = self.freduce(v)
                if ru != v and ru != rv:
                    urv = (u, rv)
                    newr.append(urv)
                    self.reductions[i] = urv
                else:
                    pass
            if len(newr) != self.reductions:
                assert ('', '') not in newr
                self.reductions = newr
            assert ('', '') not in self.reductions
            if self.confluent:
                break

        else:
            raise ValueError("""\
KnuthBendix.make_confluent did not terminate in %d iterations.
Check your equations or specify an higher max_iterations value.'
""" % max_iterations)

    def freduce(self, p):
        # This (internal) variant of reduce:
        # Uses the internal representaion:
        # Assumes p is .surrounded. by the delimiter
        # and returns the reduced value .surrounded. by it.
        # This is primarily for internal use by make_confluent

        while 1:
            q = p
            for uv in self.reductions:
                p = p.replace(*uv)
            if q == p:
                break
        return p

    def reduce(self, p):
        # This (external) variant of reduce:
        # will add delim if not .surrounded. by delim
        # but the return value will not be surrounded by it.

        if self.delim:
            p = self.wrap_delim(p)
        p = self.freduce(p)
        if self.delim:
            p = p.strip(self.delim)
        return p

    def sort(self, reds=None):
        if reds is None:
            reds = self.reductions

        def cmp(xxx_todo_changeme, xxx_todo_changeme1):
            (x, _) = xxx_todo_changeme
            (y, __) = xxx_todo_changeme1
            if self.gt(x, y):
                return 1
            if x == y:
                return 0
            return -1
        reds.sort(key=functools.cmp_to_key(cmp))

    def pp(self):
        printreds(self.reductions)

    def wrap_delim(self, p):
        if not p.startswith(self.delim):
            p = self.delim + p
        if not p.endswith(self.delim):
            p = p + self.delim
        return p


def printreds(reds):
    for i, uv in enumerate(reds):
        print('%s\t' % (uv,), end=' ')
        if (i + 1) % 4 == 0:
            print()
    if (i + 1) % 4 != 0:
        print()


def kb(E, *a, **k):
    return KnuthBendix(E, *a, **k)


class _GLUECLAMP_:
    pass


def test2():
    #
    # The group of complex numbers {1, -1, i, -i} under multiplication;
    # generators and table from Example 16.13 in (2).

    G = ['1', '-1', 'i', '-i']
    E = [('1.i',         'i'),
         ('i.i',         '-1'),
         ('i.i.i',       '-i'),
         ('i.i.i.i',     '1'),
         ]
    R = kb(E, delim='.')
    T = [['.']+G] + [[y]+[R.reduce('%s.%s' % (y, x)) for x in G] for y in G]

    assert T == [
        ['.', '1', '-1', 'i', '-i'],
        ['1', '1', '-1', 'i', '-i'],
        ['-1', '-1', '1', '-i', 'i'],
        ['i', 'i', '-i', '-1', '1'],
        ['-i', '-i', 'i', '1', '-1']]

    return R


def test():
    E = [('.a.', '.b.')]
    a = kb(E, delim='.')
    assert a('.a.', '.b.')
    E = [('fhk', 'gh'), ('m', 'kkm')]
    a = kb(E)
    p = a.reduce('fffghkkkm')
    q = a.reduce('ffghkm')
    assert p == 'ffffhm'
    assert q == 'fffhm'
    assert not a(p, q)

    E = [('.a.', '.b.')]
    a = kb(E, delim='.')
    p = a.reduce('aa')
    assert p == 'aa'
    p = a.reduce('.bb.')
    assert p == 'bb'
    p = a.reduce('b')
    assert p == 'a'

    E = [('.f.h.k.', '.g.h.'), ('.m.', '.k.k.m.')]
    a = kb(E, delim='.')
    p = a.reduce('.f.f.f.g.h.k.k.k.m.')
    q = a.reduce('.f.f.g.h.k.m.')
    assert p, q == ('.f.f.f.f.h.m.', '.f.f.f.h.m.')
    assert p == 'f.f.f.f.h.m'
    assert q == 'f.f.f.h.m'

    E = [('.f.ff.fff.', '.ffff.ff.'), ('.fffff.', '.fff.fff.fffff.')]

    a = kb(E, delim='.')

    p = a.reduce('.f.f.f.ffff.ff.fff.fff.fff.fffff.')
    q = a.reduce('.f.f.ffff.ff.fff.fffff.')

    assert p == 'f.f.f.f.ff.fffff'
    assert q == 'f.f.f.ff.fffff'


def test3():
    # From 9.3 in 251
    E = [('Hcc', 'H'),
         ('aab', 'ba'),
         ('aac', 'ca'),
         ('cccb', 'abc'),
         ('caca', 'b')]

    a = kb(E)

    canon = [
        ('Hb', 'Ha'),    ('Haa', 'Ha'),   ('Hab', 'Ha'),   ('Hca', 'Hac'),
        ('Hcb', 'Hac'),  ('Hcc', 'H'),    ('aab', 'ba'),   ('aac', 'ca'),
        ('abb', 'bb'),   ('abc', 'cb'),   ('acb', 'cb'),   ('baa', 'ba'),
        ('bab', 'bb'),   ('bac', 'cb'),   ('bba', 'bb'),   ('bca', 'cb'),
        ('bcb', 'bbc'),  ('cab', 'cb'),   ('cba', 'cb'),   ('cbb', 'bbc'),
        ('cbc', 'bb'),   ('ccb', 'bb'),   ('Haca', 'Hac'), ('Hacc', 'Ha'),
        ('bbbb', 'bb'),  ('bbbc', 'cb'),  ('bbcc', 'bbb'), ('bcca', 'bb'),
        ('caca', 'b'),   ('ccaa', 'ba'),  ('ccca', 'cb'),  ('cacca', 'cb')
    ]

    a.canon = canon

    return a



class LeftKanExtension:

    # Implementation of algorithms described by Brown and Heyworth (ref.251)
    # and Heyworth (ref.253).

    def __init__(self, mod, A, B, R, X, F):
        # External subsystem dependencies
        #    mod.KnuthBendix
        #    mod.FiniteAutomaton
        #    mod.SolveFSA
        #    mod.Cat
        #       mod.Cat.Function
        #       mod.Cat.Functor
        #       mod.Cat.check_graph
        #       mod.Cat.check_rules

        self.mod = mod
        self.Cat = mod.Cat

        #

        self.Cat.check_graph(A)
        self.Cat.check_graph(B)
        self.Cat.check_rules(R, B)

        #

        self.A = A
        self.B = B
        self.R = [(tuple(g), tuple(h)) for (g, h) in R]
        self.X = X
        self.F = F

        self.general_procedure()

    def general_procedure(self):
        self.initialize_tables()
        self.make_confluent_system()
        self.make_automaton()  # self.make_catalogue()
        self.make_natural_transformation()

    def initialize_tables(self):
        self.obj_to_str_table = {}
        self.str_to_obj_table = {}
        self.make_initial_rules()

    def make_initial_rules(self):

        # Algorithm 6.1 in (251)

        Re = []

        def add_rule(a, b):
            aw = self.make_word(a)
            bw = self.make_word(b)
            if aw != bw:
                Re.append((aw, bw))

        for a in self.A.arrows:
            srca = self.A.source(a)
            tgta = self.A.target(a)
            XA = self.X.fo(srca)
            Xa = self.X.fa(a)
            Fa = tuple(self.F.fa(a))
            Fsrca = self.F.fo(srca)
            Ftgta = self.F.fo(tgta)
            if Fa:
                t = Fsrca
                for b in Fa:
                    srcb = self.B.source(b)
                    if srcb != t:
                        raise ValueError(
                            'Arrow [%s] with source %s does not compose with target %s' % (b, srcb, t))
                    t = self.B.target(b)
                if t != Ftgta:
                    raise ValueError(
                        'Arrow %s with target %s does not compose with %s' % (Fa, t, Ftgta))
            else:
                if Fsrca != Ftgta:
                    raise ValueError(
                        'Source %s does not match target %s' % (Fsrca, Ftgta))
            for x in XA:
                add_rule(((srca, x),) + Fa, ((tgta, Xa(x)),))

        Rk = [(self.make_word(x), self.make_word(y)) for (x, y) in self.R]

        self.Re = Re
        self.Rk = Rk
        self.Rinit = Re + Rk

    def make_confluent_system(self):
        self.rs = self.mod.KnuthBendix(self.Rinit, delim='.')
        self.Rconf = self.rs.reductions

    def make_automaton(self):
        # Make nondeterministic finite automaton

        def target(e):
            if len(e) == 1 and isinstance(e[0], tuple):
                return self.F.fo(e[0][0])
            else:
                return self.B.target(e[-1])

        XA = []
        for A in self.A.objects:
            for x in self.X.fo(A):
                XA.append(((A, x),))

        follows = dict([(B, []) for B in self.B.objects])
        for b, (srcb, tgtb) in list(self.B.arrows.items()):
            follows[srcb].append((b, tgtb))

        IR = dict([(self.make_term(u), self.make_term(v))
                   for u, v in self.Rconf])

        pplR = {}
        for l, r in self.Rconf:
            t = self.make_term(l)
            for i in range(1, len(t)):
                pplR[t[:i]] = 1

        s0 = ('s0',)

        fsa = self.mod.FiniteAutomaton(s0)

        for xi in XA:
            if xi not in IR:
                fsa.add_transition(s0, xi[0], xi)

        for xi in XA:
            for b, tgtb in follows[target(xi)]:
                bterm = (b,)
                xib = xi + bterm
                if xib in pplR:
                    fsa.add_transition(xi, b, xib, tgtb)
                elif (bterm in pplR and xib not in IR):
                    fsa.add_transition(xi, b, bterm, tgtb)
                elif xib not in IR:
                    fsa.add_transition(xi, b, tgtb)

        for Bi in self.B.objects:
            for b, tgtb in follows[Bi]:
                bterm = (b,)
                if bterm in pplR:
                    fsa.add_transition(Bi, b, bterm, tgtb)
                elif bterm not in IR:
                    fsa.add_transition(Bi, b, tgtb)

        for u in pplR:
            if u in XA:
                continue
            for b, tgtb in follows[target(u)]:
                bterm = (b,)
                ub = u + bterm
                if ub in pplR:
                    fsa.add_transition(u, b, ub, tgtb)
                elif self.irreducible(ub):  # ub not in IR:
                    fsa.add_transition(u, b, tgtb)

        def get_RS(Bi):
            finals = {}
            finals[Bi] = 1
            for xi in XA:
                if self.F.fo(xi[0][0]) == Bi:
                    finals[xi] = 1
            for u in pplR:
                if target(u) == Bi:
                    finals[u] = 1

            for c in fsa.get_composites():
                for s in c:
                    if s not in finals:
                        break
                else:
                    finals[c] = 1

            dfa = fsa.get_minimized_dfa(finals)
            regexp = self.mod.SolveFSA(dfa)
            return RegularSet(regexp)

        KB = self.Cat.Function(get_RS, self.B.objects, None)

        Kb = self.Cat.Function(
            lambda a: KanAction(self.B, KB, a, target,
                                self.irreducible, self.reduce),
            self.B.arrows,
            KanAction,
        )

        self.KB = KB
        self.Kb = Kb
        self.K = self.Cat.Functor(KB, Kb)

    def make_catalogue(self):
        # Catalogue the elements of the sets pointed to by extension functor K,
        # according to algorithm described in 7.1 in (251).

        # Precondition:

        # Tables initialized and a confluent system created.
        # The system is assumed to be finite, otherwise we won't terminate.

        # Postcondition:

        # Functor self.K represented as:
        #
        # self.K.tabo = self.KB = dict mapping,
        #               source: {each B in self.B.objects}
        #               target: sets represented as lists
        # self.K.taba = self.Kb = dict, mapping
        #               source: {each a in self.B.arrows}
        #               target: tabulated function, mapping
        #                       source: KB[source of a]
        #                       target: KB[target of a]

        def target(e):
            if len(e) == 1:
                return self.F.fo(e[0][0])
            else:
                return self.B.target(e[-1])

        def add_element(e):
            if self.irreducible(e):
                block.append(e)
                KB[target(e)].append(e)
            else:
                pass

        KB = dict([(B, []) for B in self.B.objects])
        block = []

        for A in self.A.objects:
            for x in self.X.fo(A):
                add_element(((A, x),))

        while block:
            oblock = block
            block = []
            for e in oblock:
                tgt = target(e)
                for a in self.B.arrows:
                    if self.B.source(a) == tgt:
                        add_element(e + (a,))

        Kb = {}

        for a in self.B.arrows:
            src = KB[self.B.source(a)]
            tgt = KB[self.B.target(a)]
            tab = dict([(s, self.reduce(s + (a,))) for s in src])
            Kb[a] = self.Cat.Function(tab, src, tgt)

        KB = self.Cat.Function(KB, self.B.objects, list(KB.values()))
        Kb = self.Cat.Function(Kb, self.B.arrows, list(Kb.values()))

        self.KB = KB
        self.Kb = Kb
        self.K = self.Cat.Functor(KB, Kb)

    def make_natural_transformation(self):

        # Precondition:
        # initial tables should be initialized
        # self.K.fo should exist

        # Postcondition:
        #
        # self.nat[A] for A in self.A.objects

        get_nat_memo = {}

        def get_nat(A):
            if A in get_nat_memo:
                return get_nat_memo[A]

            src = self.X.fo(A)
            tgt = self.K.fo(self.F.fo(A))
            tab = dict([(x, self.reduce(((A, x),))) for x in src])
            get_nat_memo[A] = self.Cat.Function(tab, src, tgt)
            return get_nat_memo[A]

        self.nat = self.Cat.Function(get_nat, self.A.objects, None)

    def make_word(self, x):
        ots = self.obj_to_str
        return '.'.join([ots(e) for e in x if e != ''])

    def obj_to_str(self, x):
        otn = self.obj_to_str_table
        try:
            return otn[x]
        except KeyError:
            assert not (isinstance(x, tuple) and len(x) > 2)
            n = str(len(otn))
            #n = '%d:%s'%(len(otn), x)
            #n = str(x)
            otn[x] = n
            self.str_to_obj_table[n] = x
            return n

    def str_to_obj(self, x):
        return self.str_to_obj_table[x]

    def irreducible(self, x):
        tx = self.make_word(x)
        return tx == self.rs.reduce(tx)

    def reduce(self, x):
        w = self.rs.reduce(self.make_word(x))
        return self.make_term(w)

    def make_term(self, word):
        sto = self.str_to_obj_table
        return tuple([sto[s] for s in word.split('.') if s])


class KanAction:
    def __init__(self, B, KB, a, targetof, irreducible, reduce):
        srca = B.source(a)
        tgta = B.target(a)
        self.src = KB(srca)
        self.tgt = KB(tgta)
        self.a = a
        self.srca = srca
        self.targetof = targetof
        self.irreducible = irreducible
        self.reduce = reduce

    def __call__(self, s):
        if self.targetof(s) != self.srca:
            raise TypeError('''\
Target of %r (= %r) does not match source of %r (= %r)''' % (
                s, self.targetof(s), self.a, self.srca))
        if not self.irreducible(s):
            raise TypeError('''\
Argument %r is reducible to %r; and is thus not in the source set K.fo(%r)''' % (
                s, self.reduce(s), self.srca))
        return self.reduce(s + (self.a,))


class RegularSet:
    # Wraps a regular expression;
    # provides a set protocol for the underlying set of sequences:
    #  o If the RE specifies a finite language, iteration over its strings
    #  [ o set inclusion ]

    is_simplified = 0

    def __init__(self, re):
        self.re = re

    def __iter__(self):
        return iter(self.uniform)

    def __getitem__(self, x):
        return self.uniform[x]

    def __len__(self):
        return len(self.uniform)

    def get_xs_covered(self, coverage):
        N = coverage
        X = self.re.limited(coverage)
        xs = X.sequni()
        return [tuple(x) for x in xs]

    def get_uniform(self):
        self.simplify()
        return self.re.sequni()

    uniform = property(fget=get_uniform)

    def simplify(self):
        if not self.is_simplified:
            self.re = self.re.simplified()
            self.is_simplified = 1


class ObjectTester:
    def __init__(self, category_tester, object, code):
        self.category_tester = category_tester
        self.functor = category_tester.functor
        self.object = object
        self.code = code

    def get_all_arrows(self):
        return self.category_tester.arrows[self.object]

    def get_intermediate_test_code(self):
        return self.code

    def get_python_test_source_code(self):
        cmap = {
            'aseq': 'assert e[%r] == e[%r]',
            'evalfa': 'e[%r] = fa[%r](e[%r])',
            'asfo': 'assert fo[%r](e[%r])'
        }

        return '\n'.join([cmap[c[0]] % c[1:] for c in self.code])

    def execode(self, arg):
        code = self.get_python_test_source_code()

        e = {'arg': arg}
        d = {'fa': self.functor.fa,
             'fo': self.functor.fo,
             'e': e,
             }
        exec(code, d)
        return e

    def intercode(self, arg):
        e = {'arg': arg}
        fa = self.functor.fa
        fo = self.functor.fo
        for c in self.code:
            a = c[0]
            if a == 'evalfa':
                dst, ar, src = c[1:]
                e[dst] = fa[ar](e[src])
            elif a == 'asfo':
                ob, src = c[1:]
                if not fo[ob](e[src]):
                    raise ValueError('Predicate failed')
            elif a == 'aseq':
                na, nb = c[1:]
                if e[na] != e[nb]:
                    raise ValueError('e[%r] != e[%r]' % (na, nb))
            else:
                raise ValueError('Invalid code: %r' % (a,))

    def test(self, arg):
        return self.intercode(arg)


class CategoryTester:
    def __init__(self, mod, functor, arrows, get_arrow_name=None):
        self.mod = mod
        self.cat = functor.src
        self.functor = functor
        self.arrows = arrows
        if get_arrow_name is not None:
            self.get_arrow_name = get_arrow_name

    def get_arrow_name(self, a):
        return '.'.join(a)

    def get_eval_arrows_code(self, object, argname):
        fa = self.functor.fa

        name = argname
        memo = {(): name}
        memolist = [((), name)]

        codes = []

        def eval_arrow(a):
            if a in memo:
                return memo[a]
            a0 = a[:-1]
            a1 = a[-1]
            name = self.get_arrow_name(a)
            na0 = eval_arrow(a0)
            #codes.append('%s = fa[%r](%s)'%(name, a1, na0))
            codes.append(('evalfa', name, a1, na0))
            memo[a] = name
            memolist.append((a, name))
            return name

        for ar in self.arrows[object]:
            eval_arrow(ar)

        return codes, memolist

    def get_object_tester(self, object):
        code = self.get_test_object_code(object)
        return ObjectTester(self, object, code)

    def get_test_inclusion_code(self, object, ml):
        codes = []
        src = self.functor.fo.src
        for arrow, value in ml:
            ob = object
            if arrow:
                ob = self.cat.graph.target(arrow[-1])
            #codes.append('assert fo[%r](%s)'%(ob, value))
            if src is None or ob in src:
                codes.append(('asfo', ob, value))
        return codes

    def get_test_object_code(self, object):
        argname = 'arg'
        evalcodes, memolist = self.get_eval_arrows_code(object, argname)

        relcodes = self.get_test_relations_code(object, memolist)

        incodes = self.get_test_inclusion_code(object, memolist)

        return evalcodes+relcodes+incodes

    def get_test_relations_code(self, object, memolist):
        codes = []
        cat = self.cat
        fa = self.functor.fa
        memo = dict(memolist)

        def teval_arrow(ar):
            if ar in memo:
                return memo[ar]
            a0 = teval_arrow(ar[:-1])
            name = self.get_arrow_name(ar)
            #codes.append('%s = fa[%r](%s)'%(name, ar[-1], a0))
            codes.append(('evalfa', name, ar[-1], a0))
            memo[ar] = name
            return name

        # Check that the equality relations really match up
        # for all arrows in old memolist, i.e. original unique arrows
        # which is arguably overkill sometimes?..
        for a, b in cat.relations:
            a = tuple(a)
            b = tuple(b)
            src = cat.graph.source(a[0])
            for (arr, val) in memolist:
                if arr:
                    tgt = cat.graph.target(arr[-1])
                else:
                    tgt = object
                if src == tgt:
                    ara = arr + a
                    arb = arr + b
                    if ara != arb:
                        va = teval_arrow(ara)
                        vb = teval_arrow(arb)
                        assert va != vb
                        #codes.append('assert %s == %s'%(va, vb))
                        codes.append(('aseq', va, vb))
        return codes

    def test_object(self, object, value):
        tester = self.get_object_tester(object)
        tester.test(value)
        return tester

    def test_object_fail(self, object, value):
        try:
            self.test_object(object, value)
        except Exception:
            pass
        else:
            raise Exception('Exception excepted')


class _GLUECLAMP_:
    # 'imports'

    def _get_KnuthBendix(self): return self._parent.KnuthBendix.KnuthBendix
    def _get_FiniteAutomaton(self): return self._parent.FSA.FiniteAutomaton
    def _get_SolveFSA(self): return self._parent.RE.SolveFSA
    def _get_Cat(self): return self._parent.Cat

    # Main exported interface is the lke method
    # which provides a context for the LeftKanExtension class.

    def lke(self, A, B, R, X, F):
        return LeftKanExtension(self, A, B, R, X, F)

    # Other functions - examples of applications of Kan extension
    # in alphabetic order

    def arrows_map(self, cat, from_objects=0, coverage=1):
        if from_objects:
            cat = cat.get_dual()

        A = self.Cat.Graph(cat.graph.objects, [])
        B = cat.graph
        R = cat.relations
        X = self.Cat.Functor(lambda x: [1], lambda x: lambda y: y)
        F = self.Cat.Functor(lambda x: x, lambda x: [])
        ke = self.lke(A, B, R, X, F)

        memo = {}

        def get_arrows(object):
            if object in memo:
                return memo[object]
            re = ke.K.fo[object].re.rempretup()
            if from_objects:
                re = re.reversed()
            if str(coverage).startswith('length'):
                maxlen = int(coverage[6:])
                ar = []
                xs = re.get_words_memo()
                for i in range(1, maxlen+1):
                    ar.extend([tuple(x) for x in xs.get_words_of_length(i)])
            else:
                re = re.limited(coverage)
                xs = re.sequni()
                ar = [tuple(x) for x in xs]
            memo[object] = ar
            return ar

        return self.Cat.Function(
            get_arrows,
            src=ke.K.fo.src,
            tgt=None
        )

    def category_tester(self, functor, arrows=None, coverage=1):
        if isinstance(functor, tuple):
            fo, fa, src = functor
            if fo is None:
                def fo(x): return lambda y: 1
            functor = self.Cat.Functor(fo, fa, src)
        if arrows is None:
            arrows = self.arrows_map(
                functor.src, from_objects=1, coverage=coverage)
        return CategoryTester(self, functor, arrows)

    def coequalizer(self, S0, S1, f0, f1):
        # Given
        #
        # S0, S1 sets (objects that can be iterated over)
        # f0, f1 functions from S0 to S1
        #
        # Return a coequalizing function,
        # such that in the following diagram:

        #
        #  S0 ===== S0
        #  |        |
        #  | f0     | f1
        #  |        |
        #  V        V
        #  S1 ===== S1 ==== coequalizing_function.src
        #  |
        #  | coequalizing_function
        #  |
        #  V
        #  coequalizing_function.tgt

        # both paths from S0 to coequalizing_function.tgt will be equivalent,
        # and coequalizing_function.tgt is a colimit of all such sets.
        #
        # The coequalizing_function object is callable with
        # an argument from S1, and has the following attributes:
        #   .src                is identical to S1
        #   .tgt                is a set in iterable form
        #   .asdict()           returns a dict representing the mapping

        objects = [0, 1]
        arrows = {'a0': (0, 1), 'a1': (0, 1)}
        A = self.Cat.Graph(objects, arrows)

        Xo = self.Cat.Function({0: S0, 1: S1}, objects, [S0, S1])
        Xa = self.Cat.Function({'a0': f0, 'a1': f1}, arrows, [f0, f1])
        X = self.Cat.Functor(Xo, Xa)

        colimit_object, colimit_functions = self.colimit(A, X)
        return colimit_functions[1]

    def colimit(self, A, X):
        # According to 9.6 in (ref.251)

        B = self.Cat.Graph([0], {})
        R = []
        F = self.Cat.Functor(lambda x: 0, lambda x: ())

        lka = self.lke(A, B, R, X, F)

        colimit_object = lka.KB[0]
        colimit_functions = lka.nat

        # Reduce elements to a smaller (but isomorphic) form
        # I.E since elements are all of the form
        #       ((A, X),)
        # they can be reduced to the form
        #       (A, X)
        #

        colimit_object = [x[0] for x in colimit_object]

        colimit_functions = dict([
            (A, self.Cat.Function(
                dict([(a, k[0])
                      for (a, k) in list(cof.items())]),
                cof.src,
                colimit_object,
            )
            )
            for (A, cof) in list(colimit_functions.items())])

        return colimit_object, colimit_functions

    def test_arrows(self, functor, object, value):
        # Application of arrow listing to test sequencing
        # Discussed in Notes Mar 9 2005

        tester = self.category_tester(functor)
        return tester.test_object(object, value)



from tkinter import *
import tkinter.filedialog
import tkinter.messagebox

from guppy.etc.Descriptor import property_nondata


class MyVar(StringVar):
    _default = 0.0

    def set(self, value):
        StringVar.set(self, '%.2g' % value)


suffixes = ('', 'K', 'M', 'G', 'T')


def sizestring(value):
    value = float(value)
    sign = 1
    if value < 0:
        sign = -1
        value = - value
    i = 0
    while value > 99999:
        value /= 1000
        i += 1
    s = str(int(round(value)))+suffixes[i]
    if s.endswith('000'+suffixes[i]):
        s = str(int(round(value/1000)))+suffixes[i+1]
    if sign == -1:
        s = '-' + s
    return s


def percentstring(value):
    a = abs(value)
    if 10 <= a <= 9999:
        return '%d' % round(value)
    elif 0.01 <= a <= 10:
        return '%.2g' % value
    elif a <= 1e-10:
        return '0'
    else:
        return '%.0e' % value


def stringsize(s):
    if s.isdigit():
        return int(s)
    suf = s[-1:].upper()
    mult = 1000
    for su in suffixes[1:]:
        if su == suf:
            break
        mult *= 1000
    else:
        raise ValueError
    return int(s[:-1])*mult


class Menu(Menu):
    # A fix for the .delete() method in Menu.
    # To delete commands defined in the menu items deleted.
    # Also changed the comment: INDEX2 is actually INCLUDED.
    def delete(self, index1, index2=None):
        """Delete menu items between INDEX1 and INDEX2 (included)."""
        if index2 is None:
            index2 = index1
        # First find out what entries have defined commands.
        cmds = []
        for i in range(self.index(index1), self.index(index2)+1):
            c = str(self.entrycget(i, 'command'))
            if c in self._tclCommands:
                # I don't want to delete the command already, since it
                # seems mystical to do that while the entry is not yet deleted.
                cmds.append(c)
        # Delete the menu entries.
        self.tk.call(self._w, 'delete', index1, index2)
        # Now that the menu entries have been deleted,
        # we can delete their commands.
        for c in cmds:
            self.deletecommand(c)


class SizeVar(StringVar):
    _default = 0.0

    def set(self, value):
        self._value = value
        s = sizestring(value)
        StringVar.set(self, s)


class ValueLabel(Label):
    def __init__(self, *args, **kwds):
        kwds['width'] = 10
        Label.__init__(self, *args, **kwds)


class ClickButton(Button):
    # Button that runs the command directly at the click, not at release.
    # And has auto-repeat.
    def __init__(self, master, command, firstdelay=500, thendelay=150, **kwds):
        Button.__init__(self, master, **kwds)
        self._command = command
        self._firstdelay = firstdelay
        self._thendelay = thendelay
        self.bind('<Button-1>', self._event_button)
        self.bind('<ButtonRelease-1>', self._event_release)

    def _event_button(self, event=None):
        self._command()
        if event is not None:
            delay = self._firstdelay
        else:
            delay = self._thendelay
        self._after = self.after(delay, self._event_button)

    def _event_release(self, event):
        self.after_cancel(self._after)
        del self._after


class Stats:
    def __init__(self, mod, fn=None):
        self.mod = mod
        self.os = mod.os
        self.hashlib = mod.hashlib
        self.fn = fn

    def clear_cache(self):
        # It is intended to be transparently
        # automagically reopened when needed.
        self.stats = None
        del self.stats

    def get_stats(self):
        self.open(self.fn)
        return self.stats

    stats = property_nondata(get_stats)

    def collect(self):
        if not self.fn:
            return 0, 0

        stat = self.os.stat(self.fn)
        if stat == self.laststat:
            return len(self), 0

        with open(self.fn) as f:
            str = f.read(self.lastfilesize)
            md5 = self.hashlib.md5(str.encode('utf-8'))
            digest = md5.digest()
            if digest == self.lastdigest:
                numoldstats = len(self)
            else:
                self.loadstr(str, reset=1)
                numoldstats = 0

            str = f.read()

            self.laststat = self.os.fstat(f.fileno())
        self.lastfilesize = self.laststat.st_size

        md5.update(str.encode('utf-8'))
        self.lastdigest = md5.digest()
        self.loadstr(str)

        numnewstats = len(self.stats)-numoldstats

        return numoldstats, numnewstats

    def open(self, fn):
        if not fn:
            self.len_stats = 0
            self.stats = []
            self.max_size = 0
            self.fn = fn
            return

        with open(fn) as f:
            str = f.read()
            lastdigest = self.hashlib.md5(str.encode('utf-8')).digest()
            laststat = self.os.fstat(f.fileno())
        self.loadstr(str, reset=1)

        # Update these only if there was no exception so far.
        self.fn = fn
        self.lastdigest = lastdigest
        self.laststat = laststat
        self.lastfilesize = laststat.st_size

    def loadstr(self, str, reset=0):
        stats = []
        lines = str.split('\n')
        del str
        linesiter = iter(lines)
        max_size = 0
        while 1:
            try:
                st = self.mod.Use.load(linesiter)
            except StopIteration:
                break
            stats.append(st)
            if st.size > max_size:
                max_size = st.size

        # Only update self if there were no exception so far

        if reset:
            self.stats = []
            self.max_size = 0

        self.max_size = max(self.max_size, max_size)
        self.stats.extend(stats)
        self.len_stats = len(self.stats)

    def __getitem__(self, idx):
        return self.stats[idx]

    def __len__(self):
        try:
            return self.len_stats
        except AttributeError:
            self.len_stats = len(self.stats)
            return self.len_stats

    def get_max_size(self):
        return self.max_size


class ProfileRow:
    kindwidth = 30

    def __init__(self, master, row, usecolor=1):
        self.master = master
        self.row = row
        if usecolor:
            colbg = Frame(master=master, bg='black', width=1,
                          borderwidth=1, relief=GROOVE)
            self.color = Label(master=colbg, bg='white',
                               width=1, borderwidth=1, relief=GROOVE)
            self.color.grid(row=0, column=0)
            colbg.grid(row=row, column=0, sticky=NW)

        self.rsizevar = SizeVar()
        self.rsize = Label(
            master=master, textvariable=self.rsizevar, width=6, anchor=E)
        self.rpercentvar = StringVar()  # BBIntVar()
        self.rpercent = Label(
            master=master, textvariable=self.rpercentvar, width=3, anchor=E)
        self.dsizevar = SizeVar()
        self.dsize = Label(
            master=master, textvariable=self.dsizevar, width=6, anchor=E)
        self.dpercentvar = StringVar()  # BBIntVar()
        self.dpercent = Label(
            master=master, textvariable=self.dpercentvar, width=3, anchor=E)
        self.kindvar = StringVar()
        self.kind = Label(master=master, textvariable=self.kindvar, anchor=NW,
                          width=self.kindwidth, justify=LEFT)

        self.rsize.grid(row=row, column=1, sticky=NE)
        self.rpercent.grid(row=row, column=2, sticky=NE)
        self.dsize.grid(row=row, column=3, sticky=NE)
        self.dpercent.grid(row=row, column=4, sticky=NE)
        self.kind.grid(row=row, column=5, sticky=NW)

    def set_color_size_percent_kind(self, color, rsize, rpercent, dsize, dpercent, kind):
        self.set_color(color)
        if color is not None:
            self.set_color(color)
        self.rsizevar.set(rsize)
        if rpercent is None:
            rpercent = ''
        else:
            rpercent = str(int(round(rpercent)))
        self.rpercentvar.set(rpercent)

        self.dsizevar.set(dsize)
        dpercent = str(int(round(dpercent)))
        self.dpercentvar.set(dpercent)

        self.set_kind(kind)

    def set_color(self, color):
        self.color.configure(bg=color)

    def set_kind(self, kind):
        self.kindtext = kind

        if len(kind) > self.kindwidth:
            import textwrap
            kind = textwrap.fill(kind, width=self.kindwidth)
        self.kindvar.set(kind)

    def clear(self):
        self.set_color_size_percent_kind(self.master['bg'], 0, 0, 0, 0, '--')


class AxisControl:
    scale_table = [1, 2, 5]
    while scale_table[-1] < 1e12:
        scale_table.append(scale_table[-3] * 10)

    def __init__(self, master,
                 name,
                 range,
                 grid,
                 unit,
                 rangecommand,
                 gridcommand,
                 autocommand=None
                 ):
        small = 0

        self.name = name
        self.unit = unit
        self.range = range
        self.rangecommand = rangecommand

        self.frame = frame = Frame(master, borderwidth=2, relief=GROOVE)

        self.rangevar = SizeVar()
        self.rangevar.set(range)
        rangeval = Entry(master=self.frame,
                         # anchor=E,
                         width=4,
                         textvar=self.rangevar,
                         #font=('fixed', '16', 'bold'),
                         #font=('terminal', '16', 'bold'),
                         #font=('terminal', '14'),
                         font=('fixed', '14'),
                         # bg='black',fg='yellow'
                         bg='#fdd'
                         )
        rangeval.bind('<KeyPress-Return>', self.event_range_enter)

        namelabel = Menubutton(frame, text=name, relief='raised', anchor=W)

        namemenu = Menu(namelabel)
        namelabel['menu'] = namemenu

        if autocommand:
            self.autovar = BooleanVar()
            self.autovar.set(True)
            namemenu.add_checkbutton(
                # autobutton = Checkbutton(frame,
                label='Auto',
                variable=self.autovar,
                command=autocommand,
                # relief=RAISED
            )
            autobutton = Checkbutton(frame,
                                     text='Auto',
                                     variable=self.autovar,
                                     command=autocommand,
                                     relief=RAISED
                                     )

        else:
            self.autovar = None

        if gridcommand:
            self.gridvar = BooleanVar()
            self.gridvar.set(grid)
            namemenu.add_checkbutton(
                label='Grid',
                variable=self.gridvar,
                command=lambda: gridcommand(self.gridvar.get()),
            )

            gridbutton = Checkbutton(frame,
                                     text='Grid',
                                     variable=self.gridvar,
                                     command=lambda: gridcommand(
                                         self.gridvar.get()),
                                     relief=RAISED
                                     )

        rangelabel = Label(frame, text='Range')

        if name == 'Y' and small:
            padx = 5
            pady = 0
        else:
            padx = 3
            pady = 3
        ud = Frame(frame)
        rangeup = ClickButton(ud, text='+',
                              pady=pady, padx=padx,
                              font=('fixed', 8),
                              command=lambda: self.range_button(1))

        rangedown = ClickButton(ud, text='-',
                                pady=pady, padx=padx,
                                font=('fixed', 8),
                                command=lambda: self.range_button(-1))

        rangedown.grid(row=0, column=0)
        rangeup.grid(row=0, column=1)
        row = 0

        if small and name == 'Y':
            namelabel.grid(row=0, rowspan=1, column=0)
            rangeup.grid(row=0, column=1, sticky=W)

            autobutton.grid(row=1, column=0)

            rangedown.grid(row=1, column=1, sticky=W)
            rangeval.grid(row=2, column=0, columnspan=2,
                          sticky=W, padx=3, pady=3)

        elif small and name == 'X':
            namelabel.grid(row=0, column=0)
            rangeval.grid(row=0, column=1, sticky=W, padx=3, pady=3)
            rangedown.grid(row=0, column=2, sticky=W)
            rangeup.grid(row=0, column=3, sticky=W)

        else:
            namelabel.grid(row=row, column=0, sticky=N+W,
                           ipadx=0, ipady=0, padx=2, pady=2)
            rangelabel.grid(row=row, column=1, sticky=W)
            ud.grid(row=row, column=2, padx=2)
            row += 1

            if gridcommand:
                gridbutton.grid(row=row, column=0, sticky=W)

            rangeval.grid(row=row, column=1, padx=3, pady=3)
            if autocommand:
                pass
                autobutton.grid(row=row, column=2)

    def cmd_range(self):
        pass

    def event_range_enter(self, event):
        str = self.rangevar.get()
        try:
            rng = stringsize(str)
            if rng not in self.scale_table:
                if not 1 <= rng <= self.scale_table[-1]:
                    raise ValueError
        except ValueError:
            self.frame.bell()
            self.errorbox("""\
Invalid range entry.
It should be a positive integer with an optional multiplier:
K, M, G, or T
(1000, 1e6, 1e9, 1e12)
Maximum range is 1T.""")

            self.rangevar.set(self.range)
        else:
            if self.autovar:
                self.autovar.set(False)
            self.setrange(rng)

    def auto_command(self):
        pass

    def errorbox(self, msg):
        tkinter.messagebox.showerror(master=self.frame, message=msg)

    def fit(self, range):
        range = self.scale_by_table(range)
        self.setrange(range)

    def range_button(self, d):
        if self.autovar:
            self.autovar.set(False)
        self.range_change(d)

    def range_change(self, d):
        range = self.range
        srange = self.scale_by_table(range)
        if srange > range:
            if d > 0:
                d -= 1
        i = self.scale_table.index(srange)
        i += d
        if i >= len(self.scale_table):
            i = len(self.scale_table) - 1
        if i < 0:
            i = 0

        self.setrange(self.scale_table[i])

    def setrange(self, range):
        if range != self.range:
            self.range = range
            self.rangevar.set(range)
            self.rangecommand(range)

    def scale_by_table(self, s):
        # Return the scale from table that is higher or equal to s
        for ts in self.scale_table:
            if ts >= s:
                return ts
        return self.scale_table[-1]


WM = 1


class Marker:
    def __init__(self, d, tag, name, pos, poscommand=None):
        self.d = d
        self.tag = tag
        self.name = name
        self.xmarker = pos
        self.butdown = 0
        self.ocursor = d.ocursor
        self.cursor = self.ocursor
        self.poscommand = None
        self.intpos = None
        self.moving = 0
        self.selected = 0
        self.entered = 0
        self.butdownselected = 0
        self.motion_id = None
        self.create()

    def bind(self, sequence, function):
        tag = self.tag
        self.d.drawingarea.tag_bind(tag, sequence, function)
        if WM:
            self.xlabel.bind(sequence, function)
        else:
            self.d.xmarks.tag_bind(tag, sequence, function)

    def coords(self, canx):
        self.d.drawingarea.coords(self.tag,
                                  canx, 0,
                                  canx, -int(self.d.boty))

        self.d.xmarks.coords(self.tag, canx, 10)

    def create(self):
        tag = self.tag
        text = self.name
        pos = 0

        self.d.drawingarea.create_line(pos, 0, pos, 20-self.d.boty, stipple='gray12',
                                       width=4, tags=(tag,))
        if WM:
            label = self.xlabel = Label(
                self.d.xmarks, text=text, padx=2, pady=2, relief=RAISED)
            self.d.xmarks.create_window(pos, 0, window=label, tags=(tag,))
        else:
            self.d.xmarks.create_text(pos, 0, text=text, tags=(tag,))

        self.bind('<Button-1>', self.event_button_1)
        self.bind('<ButtonRelease-1>', self.event_button_1_release)
        self.bind('<Enter>', self.event_enter)
        self.bind('<Leave>', self.event_leave)

        self.d.drawingarea.bind('<Enter>', self.event_enter_movearea, add='+')
        self.d.drawingarea.bind(
            '<Button-1>', self.event_button_1_movearea, add='+')

    def event_button_1(self, event):
        self.butdown = 1
        if self.selected:
            self.butdownselected = 1
            if self.moving:
                self.event_stop_move(event)
        else:
            self.butdownselected = 0

        self.has_moved = 0
        self.event_selected(event)
        self.event_start_move(event)

    def event_button_1_movearea(self, event):
        if not self.entered:
            self.event_deselected(event)

    def event_button_1_release(self, event):
        self.butdown = 0
        if self.has_moved == self.butdownselected:
            if self.selected:
                if self.moving and not (self.disloy <= event.y_root < self.dishiy):
                    self.event_stop_move(None)
                    self.setcursor(self.ocursor)
            else:
                self.setcursor(self.ocursor)
            return
        self.event_deselected(event)

    def event_deselected(self, event):
        if self.selected:
            self.selected = 0
            self.xlabel['relief'] = RAISED
            if self.moving:
                self.event_stop_move(event)

    def event_enter(self, event):
        self.entered = 1
        if not self.moving:
            if self.selected:
                self.event_start_move(event)
            else:
                self.setcursor('hand2')

    def event_enter_movearea(self, event):
        if self.selected and not self.moving:
            self.event_start_move(event)

    def event_leave(self, event):
        self.entered = 0
        if not self.moving:
            self.setcursor(self.ocursor)
        elif not (self.fraloy <= event.y_root < self.frahiy):
            pass

    def event_motion(self, event):
        self.has_moved = 1

        inside = (self.fraloy <= event.y_root < self.frahiy)

        if inside != self.inside:
            self.inside = inside
            if not inside:
                self.out_event = event
                self.event_stop_move(None)
                if self.butdown:
                    self.setcursor('circle')
                    self.d.bind_motion(self.event_motion_downout)
            else:
                self.in_event = event
                #self.delta += self.out_event.x_root - event.x_root
                self.event_start_move(event)
                return

        if inside:
            self.moved(event)
            self.setxvars()

    def event_motion_downout(self, event):
        # We don't get an enter while button is pressed down
        # Emulate an enter if we detect entering
        inside = (self.fraloy <= event.y_root < self.frahiy)
        if inside:
            self.d.unbind_motion(self.event_motion_downout)
            self.event_enter_movearea(event)

    def event_selected(self, event):
        for m in self.d.marks:
            m.event_deselected(event)
        self.selected = 1
        self.xlabel['relief'] = SUNKEN

    def event_start_move(self, event):
        self.moving = 1
        self.fralox = self.d.frame.winfo_rootx()
        self.frahix = self.fralox + self.d.frame.winfo_width()

        self.fraloy = self.d.frame.winfo_rooty()
        self.frahiy = self.fraloy + self.d.frame.winfo_height()

        self.dislox = self.d.drawingarea.winfo_rootx()
        self.dishix = self.dislox + self.d.drawingarea.winfo_width()
        self.disloy = self.d.drawingarea.winfo_rooty()
        self.dishiy = self.disloy + self.d.drawingarea.winfo_height()

        self.down_event = event
        self.prev_event = event
        self.down_xmarker = self.xmarker
        self.down_xvfrac = self.d.drawingarea.xview()[0]
        self.inside = 1
        self.delta = 0
        self.lift()

        self.motion_id = self.d.bind_motion(self.event_motion)
        self.moved(event)

    def event_stop_move(self, event):
        assert self.moving
        self.moving = 0

        self.d.unbind_motion(self.motion_id)
        if event is not None:
            self.moved(event)
            self.setxvars()

        if self.entered and not self.selected:
            self.setcursor('hand2')
        else:
            self.setcursor(self.ocursor)

    def lift(self):
        self.d.xmarks.tag_raise(self.tag)
        if WM:
            self.xlabel.lift()
        self.d.drawingarea.tag_raise(self.tag)

    def move(self, sample):
        canx = self.d.canxscaled(sample)
        self.d.xview_pos(canx)
        self.coords(canx)
        self.xmarker = sample
        self.lift()

    def moved(self, event):
        curx = event.x_root
        cury = event.y_root
        prevx = self.prev_event.x_root

        if prevx > self.dishix and curx < self.dishix:
            prevx = self.dishix
        elif prevx < self.dislox and curx > self.dislox:
            prevx = self.dislox

        markx = self.d.canxscaled(self.xmarker) - \
            self.d.drawingarea.canvasx(0) + self.dislox

        dx = curx - prevx
        l = r = 1

        if self.xmarker >= self.d.numstats-1:
            r = 0

        if self.xmarker <= 0:
            l = 0

        stop = 0

        # Should we allow to move it back or not
        # if it is at an endpoint?
        # Here we don't move it at all, to make marker pos correspond
        # more closely with mouse position.

        if ((r == 0 and curx > markx) or (l == 0 and curx < markx)):
            l = r = 0

        if self.butdown:
            if curx > self.dishix:
                l = 0
            elif curx < self.dislox:
                r = 0
        else:
            if not (self.dislox <= curx < self.dishix and
                    self.disloy <= cury < self.dishiy):
                l = r = 0
                stop = 1

        if l and r:
            self.setcursor('sb_h_double_arrow')
        elif l:
            self.setcursor('sb_left_arrow')
            if dx > 0:
                dx = 0
        elif r:
            self.setcursor('sb_right_arrow')
            if dx < 0:
                dx = 0
        else:
            self.setcursor('dot')
            dx = 0

        self.prev_event = event

        sample = self.d.limitx(self.xmarker + dx / self.d.xscale)
        canx = self.d.canxscaled(sample)
        self.d.xview_pos(canx)
        self.coords(canx)
        self.xmarker = sample
        if stop and self.moving:
            self.event_stop_move(None)

    def set(self):
        canx = self.d.canxscaled(self.xmarker)
        self.coords(canx)
        self.lift()

    def set_poscommand(self, command):
        self.poscommand = command
        self.intpos = None

    def setcursor(self, cursor):
        if cursor != self.cursor:
            self.xlabel['cursor'] = cursor
            self.cursor = cursor
        self.d.setcursor(cursor)

    def setxvars(self):
        if self.poscommand:
            intpos = int(round(self.xmarker))
            if intpos != self.intpos:
                self.intpos = intpos
                self.poscommand(intpos)


class Display:
    orgwidth = 300
    orgheight = 300
    minwidth = 30
    minheight = 30

    def __init__(self, master,
                 scale_table,
                 numkindrows,
                 getkindcolor,
                 xrange=100,
                 yrange=100,
                 xgrid=False,
                 ygrid=False,
                 graphtype='Bars',
                 statype='Size',
                 ):
        self.master = master
        self.scale_table = scale_table
        self.numkindrows = numkindrows
        self.getkindcolor = getkindcolor
        self.xrange = xrange
        self.yrange = yrange
        self.xgrid = xgrid
        self.var_xgrid = BooleanVar(xgrid)
        self.var_xgrid.set(xgrid)
        self.var_ygrid = BooleanVar(xgrid)
        self.ygrid = ygrid
        self.var_ygrid.set(ygrid)
        self.graphtype = graphtype
        self.statype = statype

        self.numstats = 0
        self.ymaxs = []
        self.ymins = []
        self.ymax = 1

        # To get around problems with dynamic unbinding / unbinding of motion,
        # I handle it myself. in the bind_motion method using the following.
        self.bound_motions = {}
        self.event_motion_id = None
        #

        self.frame = frame = Frame(master,
                                   borderwidth=3,
                                   relief=SUNKEN,
                                   # relief=GROOVE,
                                   # background='green'
                                   )
        #self.frame = frame = Frame(master,background='green')

        bordercolor = '#ccc'
        screencolor = '#e0e0e0'
        xscrollincrement = 1
        frame = Frame(self.frame)
        frame.grid(row=0, column=0)
        #move = Frame(frame, height=10,width=10,background='red', relief=RAISED)
        #move = Button(self.frame, height=10,width=10,background='red')
        self.drawingarea = C = Canvas(frame,
                                      width=self.orgwidth,
                                      height=self.orgheight,
                                      xscrollincrement=xscrollincrement,
                                      # background='black',
                                      background=screencolor,
                                      bd=0,
                                      xscrollcommand=self.xscrollbar_set,
                                      # confine=False,
                                      )

        #self.yctrlframe = Frame(frame, borderwidth=2,relief=GROOVE)
        self.yscrollbar = Scrollbar(frame, orient=VERTICAL, width=10)
        # self.yscrollbar['command']=self.drawingarea.yview
        #self.drawingarea['yscrollcommand'] = self.yscrollbar_set
        # self.yscrollbar.pack(side=RIGHT,fill=Y)
        #self.yctrlframe.grid(row = 0, column = 0,sticky=N+S,padx=3,pady=3)

        self.xaxis = Canvas(frame,
                            width=C['width'],
                            height=20,
                            xscrollincrement=xscrollincrement,
                            bd=0,
                            background=bordercolor,
                            #xscrollcommand = self.xscrollbar_set
                            # confine=False,
                            )
        self.xmarks = Canvas(frame,
                             width=C['width'],
                             height=20,
                             xscrollincrement=xscrollincrement,
                             bd=0,
                             background=bordercolor,
                             #xscrollcommand = self.xscrollbar_set
                             # confine=False,
                             )
        self.yaxis = Canvas(frame, height=C['height'], width=50,
                            bd=0,
                            background=bordercolor,
                            )

        self.xscrollbar = Scrollbar(frame, orient=HORIZONTAL,
                                    command=self.drawingarea_xview,
                                    width=12,
                                    background=bordercolor,

                                    )

        xy = Canvas(frame, width=50, height=20, bd=0,
                    background=bordercolor,
                    )

        var_yrange = SizeVar()
        self.var_yrange = var_yrange

        row = 0

        Label(frame,
              textvar=var_yrange,
              bd=0,
              relief=FLAT,
              background=bordercolor).grid(
                  row=row,
                  column=0,
                  sticky=W+E+N+S)

        self.xscrollbar.grid(row=row, column=1, sticky=E+W)

        row += 1

        self.yunit = Label(frame,
                           text='Bytes',
                           bd=0,
                           relief=FLAT,
                           background=bordercolor)
        self.yunit.grid(
            row=row,
            column=0,
            sticky=W+E+N+S)

        self.xmarks.grid(row=row, column=1, sticky=W+E+N)

        row += 1

        self.yaxis.grid(row=row, column=0)
        C.grid(row=row, column=1, sticky=W+E)

        row += 1

        xy.grid(row=row, column=0)
        self.xaxis.grid(row=row, column=1, sticky=W+E+N)

        self.botx = float(C['width'])
        self.boty = float(C['height'])
        self.chdim = self.getchdim()
        self.canx0 = 0
        self.tmax = 0
        self.xscale = self.botx / self.xrange
        self.yscale = self.boty / self.yrange
        self.xi0 = None

        xy.create_line(0, 2, 44, 2)
        xy.create_line(49, 6, 49, 22)
        xy.create_text(25, 14, text='Sample')

        self.setscrollregion()

        self.ocursor = self.drawingarea['cursor']
        self.cursor = self.ocursor

        self.marks = []

    def bind_motion(self, function):
        if self.event_motion_id == None:
            self.event_motion_id = self.frame.bind_all(
                '<Motion>', self.event_motion, add='+')
        self.bound_motions[function] = self.bound_motions.get(function, 0) + 1
        return function

    def event_motion(self, event):
        for f in list(self.bound_motions.keys()):
            f(event)

    def unbind_motion(self, funcid):
        n = self.bound_motions[funcid] - 1
        if n == 0:
            del self.bound_motions[funcid]
        else:
            self.bound_motions[funcid] = n

    def new_xmarker(self, name=None, pos=0):
        tag = 'M%d' % len(self.marks)
        if name is None:
            name = tag
        m = Marker(self, tag, name, pos)
        self.marks.append(m)
        return m

    def canxscaled(self, x):
        return x * self.xscale + self.canx0

    def canyscaled(self, y):
        return - y * self.yscale

    def cmd_xgrid(self):
        self.xgrid = self.var_xgrid.get()
        self.drawxaxis()

    def cmd_ygrid(self):
        self.ygrid = self.var_ygrid.get()
        self.drawyaxis()

    def cmd_yrange_auto(self):
        self.ymax = None
        self.yrange_auto()

    def limitx(self, x):
        lo = 0
        hi = max(0, self.numstats-1)
        if x < lo:
            return lo
        if x > hi:
            return hi
        return x

    def resize(self, dx, dy):
        x = self.botx + dx
        y = self.boty + dy
        if x < self.minwidth:
            x = self.minwidth
            dx = x - self.botx
        if y < self.minheight:
            y = self.minheight
            dy = y - self.boty

        xv = self.drawingarea.xview()
        yv = self.drawingarea.yview()

        self.drawingarea.configure(width=x, height=y)
        self.xaxis.configure(width=x)
        self.xmarks.configure(width=x)
        self.yaxis.configure(height=y)

        xscale = float(x) / self.xrange
        yscale = float(y) / self.yrange

        xscaleorg = self.drawingarea.canvasx(0)
        yscaleorg = 0

        xq = xscale / self.xscale
        yq = yscale / self.yscale

        self.drawingarea.scale("all", xscaleorg, yscaleorg, xq, yq)
        #self.drawingarea.scale("barsep",xscaleorg, yscaleorg, xq, yq)
        #self.drawingarea.scale("xmarker",xscaleorg, yscaleorg, xq, yq)

        self.canx0 = xscaleorg + (self.canx0 - xscaleorg) * xq

        self.botx = x
        self.boty = y
        self.xscale = xscale
        self.yscale = yscale

        self.drawxaxis()
        self.drawyaxis()

        self.setscrollregion()

        # If the size changed much, the canvas may scroll though it shouldn't.
        # Notes 11 and 26 Oct 2005 .
        # I save the current scroll position.
        # The caller has to call the .moveback() method some time later.
        self.wantedpos = xv[0]

        return dx, dy

    def moveback(self):
        self.frame.update_idletasks()
        self.xview(MOVETO, self.wantedpos)

    def draw():
        self.drawxaxis()
        self.drawyaxis()

    def draw_stat(self, idx, stat):
        graphtype = self.graphtype
        statype = self.statype

        rows = stat.get_rows_n_and_other(self.numkindrows, statype)
        if statype == 'Size':
            kindval = dict([(r.name, r.size) for r in rows])
        else:
            kindval = dict([(r.name, r.count) for r in rows])
        order = [r.name for r in rows]

        order.reverse()

        lastkindval = self.lastkindval
        self.lastkindval = kindval

        C = self.drawingarea

        yscale = self.yscale
        xscale = self.xscale

        x0 = idx * xscale - 0.5 * xscale + self.canx0
        x1 = x0 + xscale
        ymax = 0
        ymin = 0

        y = 0

        bw = 0.05*xscale

        ocolor = None
        for k in order:
            dy = kindval.get(k, 0)
            if not dy:
                continue
            color = self.getkindcolor(k)

            if graphtype == 'Bars':
                line = C.create_rectangle(x0+bw, -y*yscale,
                                          x1-bw, -(y+dy)*yscale,
                                          fill=color,
                                          outline=color,
                                          width=0,
                                          tags=("a",))
                if color == ocolor:
                    C.create_line(x0, -(y)*yscale,
                                  x1, -(y)*yscale,
                                  fill='black',
                                  tags=('barsep',))
                ocolor = color
                y += dy
            elif graphtype == 'Lines':
                if dy > ymax:
                    ymax = dy
                elif dy < ymin:
                    ymin = dy
                y0 = lastkindval.get(k)
                if y0 is None:
                    y0 = dy
                    x00 = x0
                else:
                    x00 = x0 - 0.4 * xscale
                    C.create_line(x00,  - y0 * yscale,
                                  x1 - 0.6 * xscale,  - dy * yscale,
                                  fill=color,
                                  tags=('a',))

                C.create_line(x1 - 0.6 * xscale,  - dy * yscale,
                              x1 - 0.4 * xscale,  - dy * yscale,
                              fill=color,
                              width=4,
                              tags=('a',))

        if graphtype == 'Bars':
            if y > ymax:
                ymax = y
            elif y < ymin:
                ymin = y

        assert idx == len(self.ymaxs) == len(self.ymins)
        self.ymaxs.append(ymax)
        self.ymins.append(ymin)

        if idx > self.tmax:
            self.tmax = idx

    def drawingarea_xview(self, cmd, what, unit=None):
        if cmd == 'scroll' and unit == 'units':
            what = int(max(2, self.xscale)*int(what))

        self.xview(cmd, what, unit)

    def setcursor(self, cursor):
        if cursor != self.cursor:
            self.drawingarea['cursor'] = cursor
            self.master['cursor'] = cursor
            self.cursor = cursor

    def xmarkers_set(self):
        for m in self.marks:
            m.set()

    def xview(self, *args):
        if not args:
            return self.drawingarea.xview()
        self.drawingarea.xview(*args)
        self.xaxis.xview(*args)
        self.xmarks.xview(*args)

    def xview_moveto(self, fraction):
        self.xview(MOVETO, fraction)

    def xview_pos(self, pos, fraction=None, leftmargin=5, rightmargin=5):
        # Scroll canvas view, if necessary, so that something
        # (eg an x marker) at canvas position pos will be visible
        # with minimum specified margin at left and right.
        # Scroll relative to fraction; default is current xview position.

        if fraction is None:
            fraction = self.xview()[0]

        x1, y1, x2, y2 = self.scrollregion

        cc = x1 + fraction * (x2 - x1)

        xm = pos - cc

        lo = leftmargin
        hi = self.botx - rightmargin

        if xm < lo:
            dx = xm - lo
            xm = lo
        elif xm >= hi:
            dx = (xm - hi)
            xm = hi
        else:
            dx = 0

        r = fraction + dx / float(x2 - x1)
        self.xview_moveto(r)

    def drawxaxis(self):
        scale_table = self.scale_table
        self.xaxis.delete('all')
        self.drawingarea.delete('xgrid')
        x1, y1, x2, y2 = self.scrollregion

        chdx, chdy = self.chdim

        i = 0
        while (scale_table[i] * self.xscale <
               min(5, len(str(scale_table[i] * self.tmax))) * chdx):
            i += 1
        self.xstep = scale_table[i]

        divisuf = (
            (1000000000000, '%dT'),
            (1000000000, '%dG'),
            (1000000,  '%dM'),
            (1000,  '%dK'),
            (1, '%d')
        )

        for divi, form in divisuf:
            if self.xstep >= divi:
                break

        self.xdivi = divi
        self.xform = form

        self.xi0 = 0
        self.updatexaxis()

    def updatexaxis(self):

        chdx, chdy = self.chdim
        step = self.xstep

        gridon = self.xgrid

        for i in range(self.xi0, self.tmax+step, step):
            x = self.canx0 + i*self.xscale
            self.xaxis.create_line(x, 0, x, 4)
            if gridon:
                self.drawingarea.create_line(x, 0, x, -self.boty,
                                             tags=('xgrid',), width=2, stipple="gray25")
            text = self.xform % (i / self.xdivi)
            self.xaxis.create_text(x, chdy,  text=text)

        self.xaxis.create_line(self.canx0 + self.xi0 *
                               self.xscale, 1, x+self.xscale, 1)

        self.xi0 = i
        self.xmarkers_set()

    def drawyaxis(self):

        gridon = self.ygrid

        self.yaxis.delete('all')
        self.drawingarea.delete('ygrid')

        chdx, chdy = self.getchdim()

        width = int(self.yaxis['width'])
        i = 0
        maxval = self.yrange

        while (self.scale_table[i] * self.yscale < 1.5 * chdy):
            i += 1
        step = self.scale_table[i]

        divisuf = (
            (1000000000000, '%4dT'),
            (1000000000, '%4dG'),
            (1000000,  '%4dM'),
            (1000,  '%4dK'),
            (1, '%5d')
        )

        for divi, form in divisuf:
            if step >= divi:
                break

        for i in range(0, maxval+step, step):
            y = - i*self.yscale
            self.yaxis.create_line(width-3, y, width-1, y)
            if gridon:
                self.drawingarea.create_line(self.scrollregion[0], y,
                                             self.scrollregion[2], y,
                                             stipple="gray25",
                                             tags=('ygrid',))
            text = form % (i / divi)
            self.yaxis.create_text(chdx*2.5, y-0.5*chdy,  text=text)

        #self.yaxis.create_text(chdx*2.5, 0.5*chdy, text='bytes')

        self.yaxis.create_line(width-1, 0, width-1, -self.boty)
        self.xmarkers_set()

    def getchdim(self):
        ch = self.xaxis.create_text(0, 0, text='0')
        x1, y1, x2, y2 = self.xaxis.bbox(ch)
        self.xaxis.delete(ch)
        chdx = abs(x2 - x1)
        chdy = abs(y2 - y1)

        return chdx, chdy

    def load_stats(self, stats):

        ocursor = self.frame.winfo_toplevel()['cursor']
        try:
            self.frame.winfo_toplevel()['cursor'] = 'watch'
            self.frame.update()

            self.numstats = len(stats)

            self.lastkindval = {}
            self.tmax = 0
            self.ymax = None
            self.ymaxs = []
            self.ymins = []

            C = self.drawingarea

            C.delete('barsep')
            C.delete('a')

            for (i, st) in enumerate(stats):
                self.draw_stat(i, st)

            try:
                self.drawingarea.tag_raise('barsep', 'a')
            except TclError:
                pass  # May be 'tagOrId "a" doesn't match any items' if empty!

            self.drawxaxis()
            self.drawyaxis()
            self.xmarkers_set()
            self.yrange_auto()
        finally:
            self.frame.winfo_toplevel()['cursor'] = ocursor

    def add_stats(self, stats):
        for (i, st) in enumerate(stats):
            self.draw_stat(i+self.numstats, st)
        self.numstats += len(stats)
        self.updatexaxis()
        self.setscrollregion()

    def setxgrid(self, grid):
        self.xgrid = grid
        self.drawxaxis()

    def setygrid(self, grid):
        self.ygrid = grid
        self.drawyaxis()

    def setgraphtype(self, gmode, stats):
        graphtype, statype = gmode.split(' ')
        if graphtype != self.graphtype or statype != self.statype:
            self.graphtype = graphtype
            self.statype = statype
            if statype == 'Size':
                self.yunit['text'] = 'Bytes'
            elif statype == 'Count':
                self.yunit['text'] = 'Objects'
            else:
                raise ValueError
            self.load_stats(stats)

    def setscrollregion(self):
        C = self.drawingarea
        botx = self.botx
        x1 = self.canx0
        x2 = self.tmax * self.xscale + self.canx0
        x1extra = botx / 2 + 2  # max(5, self.xscale*0.5)
        x2extra = botx / 2 + 2  # max(5, self.xscale*0.5)

        x1 -= x1extra
        x2 += x2extra

        y1 = 1-self.boty
        y2 = 1

        self.scrollregion = (x1, y1, x2, y2)
        C.configure(scrollregion=self.scrollregion)

        self.xaxis.configure(scrollregion=(x1, 0, x2, 10))
        self.xmarks.configure(scrollregion=(x1, 0, x2, 20))
        self.yaxis.configure(scrollregion=(0, y1, 20, y2))

        self.drawingarea.yview(MOVETO, 0.0)

    def setxrange(self, xrange):
        dxrange = self.xrange / float(xrange)
        self.xrange = xrange
        xscaleorg = self.drawingarea.canvasx(self.botx/2)
        self.drawingarea.scale("a", xscaleorg, 0, dxrange, 1.0)
        self.drawingarea.scale("barsep", xscaleorg, 0, dxrange, 1.0)
        self.canx0 = xscaleorg + (self.canx0 - xscaleorg) * dxrange
        self.xscale = self.botx / float(self.xrange)
        self.setxscrollincrement(max(2, self.xscale))
        self.drawxaxis()
        self.setscrollregion()

    def setxscrollincrement(self, dx):
        return
        self.drawingarea.configure(xscrollincrement=dx)
        self.xaxis.configure(xscrollincrement=dx)
        self.xmarks.configure(xscrollincrement=dx)

    def setyrange(self, yrange):

        dyrange = float(self.yrange) / yrange
        self.yrange = yrange
        self.var_yrange.set(yrange)

        self.drawingarea.scale("a", 0, 0, 1.0, dyrange)
        self.drawingarea.scale("barsep", 0, 0, 1.0, dyrange)
        self.yscale = float(self.boty) / self.yrange
        self.drawingarea.yview(MOVETO, 0.0)
        self.drawyaxis()

    def xscrollbar_set(self, first, last):
        self.xscrollbar.set(first, last)
        self.yrange_auto()

    def yrange_auto(self, force=0):
        if force or self.ycontrol.autovar.get():
            lo = max(0,
                     int(0.5+(self.drawingarea.canvasx(0) - self.canx0) / self.xscale))
            hi = min(len(self.ymaxs),
                     int(1.5+(self.drawingarea.canvasx(self.botx) - self.canx0) / self.xscale))
            if lo == hi:
                ymax = 1
            else:
                ymax = max(self.ymaxs[lo:hi])
            if ymax != self.ymax:
                self.ymax = ymax
                self.ycontrol.fit(ymax)


class MarkerControl:
    def __init__(self, master,
                 marker,
                 setcommand=lambda: 0
                 ):
        self.sample = 0
        self.numsamples = 0
        self.setcommand = setcommand
        self.marker = marker
        self.name = marker.name
        sf = self.frame = Frame(master, borderwidth=2, relief=GROOVE)
        self.samplevar = SizeVar()
        Label(sf, text='%s sample' % marker.name).grid(row=0, column=0)
        Label(sf,
              textvariable=self.samplevar,
              font=('terminal', '16', 'bold'),
              bg='black', fg='yellow'
              ).grid(row=1, column=0, padx=3, pady=3)
        ClickButton(sf, text='-',
                    pady=0, padx=5,
                    command=lambda: self.changesample(-1)).grid(row=0, column=1, sticky=E)
        ClickButton(sf, text='+',
                    pady=0, padx=5,
                    command=lambda: self.changesample(1)).grid(row=0, column=2, sticky=W)
        self.trackingvar = BooleanVar()
        self.trackbutton = Checkbutton(
            sf, text='Track',
            padx=5,

            variable=self.trackingvar,
            relief=RAISED,
            command=self.settracking,
            indicatoron=1,
        )
        self.trackbutton.grid(row=1, column=1, columnspan=2)

    def changesample(self, d):
        sample = self.sample + d
        if 0 <= sample < self.numsamples:
            self.setmarker(sample)

    def setmarker(self, sample):
        self.marker.move(sample)
        self.setsample(sample)

    def setnumsamples(self, num):
        self.numsamples = num
        if self.trackingvar.get() or self.sample >= self.numsamples:
            self.setmarker(max(0, self.numsamples-1))

    def setsample(self, sample):
        self.sample = sample
        self.samplevar.set(sample)
        self.setcommand()

    def settracking(self, tracking=None):
        if tracking is not None:
            self.trackingvar.set(tracking)
        else:
            tracking = self.trackingvar.get()
        if tracking:
            self.setmarker(max(0, self.numsamples-1))


class Window:
    def __init__(self, app, frame, windowmenu=None):
        self.app = app
        self.frame = frame
        self.windowmenu = windowmenu
        self.wtitle = frame.title()
        self._is_destroyed = 0
        # Binding to <destroy> didnt work well:
        #   frame.bind('<Destroy>', self.event_destroy, add='+')
        # I give up. I modify .destroy of frame argument instead.
        self.old_destroy = frame.destroy
        frame.destroy = self.new_destroy

    def new_destroy(self):
        if self._is_destroyed:
            return
        self._is_destroyed = 1
        self.app.del_window(self)
        try:
            self.old_destroy()
        except TclError:
            # This may happen at closing last window
            # because exit destroys the root when it sees all windows were closed.
            # So I ignore it.
            pass

    def title(self, title):
        self.frame.title(title)
        self.frame.iconname(title)
        self.wtitle = title
        self.app.chg_window(self)

    def wakeup(self):
        frame = self.frame
        try:
            if frame.wm_state() == "iconic":
                frame.wm_deiconify()
            frame.tkraise()
            # I don't think I want .focus_set: it behaved strange in X at least.
            # frame.focus_set()
        except TclError:
            # This can happen when the window menu was torn off.
            # Simply ignore it.
            pass


class WindowMenu:
    def __init__(self, frame, variable):
        self.button = Menubutton(frame, text='Window')
        self.menu = Menu(self.button)
        self.button['menu'] = self.menu
        self.variable = variable
        self.wmap = {}

    def add_window(self, window):
        self.menu.add_radiobutton(
            command=window.wakeup,
            label='%d %s' % (window.wid, window.wtitle),
            value=window.wid,
            variable=self.variable)
        self.wmap[window.wid] = self.menu.index(END)

    def chg_window(self, window):
        self.menu.delete(self.wmap[window.wid])
        self.menu.insert_radiobutton(
            self.wmap[window.wid],
            command=window.wakeup,
            label='%d %s' % (window.wid, window.wtitle),
            value=window.wid,
            variable=self.variable)

    def del_window(self, window):
        idx = self.wmap[window.wid]
        del self.wmap[window.wid]
        try:
            self.menu.delete(idx)
        except TclError:
            # This can happen if the menu was destroyed before its contents.
            # Simply ignore it.
            pass
        for wid in list(self.wmap.keys()):
            if self.wmap[wid] > idx:
                self.wmap[wid] -= 1


class ProfileApp:
    def __init__(self, mod):
        self.mod = mod
        root = Tk()
        self.root = root
        root.withdraw()
        self.windows = {}
        self.windowmenus = {}
        self.var_window = IntVar(root)

    def add_window(self, window):
        window.wid = max([0]+list(self.windows.keys()))+1
        self.windows[window.wid] = window
        wm = getattr(window, 'windowmenu', None)
        if wm:
            self.windowmenus[window.wid] = wm
            for w in list(self.windows.values()):
                if w is not window:
                    wm.add_window(w)
        for wm in list(self.windowmenus.values()):
            wm.add_window(window)
        self.var_window.set(window.wid)
        window.frame.bind('<FocusIn>',
                          lambda event: self.var_window.set(window.wid), add='+')
        window.frame.bind('<Deactivate>',
                          lambda event: self.var_window.set(0), add='+')

    def add_window_frame(self, frame, windowmenu=None):
        w = Window(self, frame, windowmenu)
        self.add_window(w)
        return w

    def chg_window(self, window):
        for wm in list(self.windowmenus.values()):
            wm.chg_window(window)

    def del_window(self, window):
        wid = window.wid
        if getattr(window, 'windowmenu', None):
            del self.windowmenus[wid]
        del self.windows[wid]
        for wm in list(self.windowmenus.values()):
            wm.del_window(window)
        if not self.windows:
            self.exit()

    def exit(self):
        try:
            self.root.destroy()
        except TclError:
            pass
        self.root.quit()

    def mainloop(self):
        return self.root.mainloop()

    def new_profile_browser(self, filename):
        return ProfileBrowser(self, filename)


class PaneDiv:
    def __init__(self, master, movecommand):
        self.frame = frame = Frame(master)
        self.movecommand = movecommand

        self.butsize = bs = 6
        bc = self.butcent = bs / 2 + 3
        h = 10
        self.top = Canvas(
            frame,
            width=10,
            height=h,
        )

        self.top.create_line(
            bc, 0, bc, h, fill='#808080', width=1)
        self.top.create_line(
            bc+1, 0, bc+1, h, fill='white', width=1)

        self.rsbut = Canvas(
            frame,
            cursor='crosshair',
            width=self.butsize,
            height=self.butsize,
            relief=RAISED,
            bd=2
        )

        self.bot = Canvas(
            frame,
            width=10,
            height=300,
            bd=0
        )

        self.top.grid(row=0, column=0, sticky=N)
        self.rsbut.grid(row=1, column=0, sticky=N)
        self.bot.grid(row=2, column=0, sticky=N)

        self.rsbut.bind('<Button-1>', self.but_down)
        self.rsbut.bind('<ButtonRelease-1>', self.but_up)

    def but_down(self, event):
        self.down_event = event
        self.rsbut.configure(relief=SUNKEN)

    def but_up(self, event):
        self.rsbut.configure(relief=RAISED)
        dx = event.x - self.down_event.x
        self.movecommand(dx)

    def setheight(self, height):
        h = height - 18
        self.bot['height'] = h

        bc = self.butcent

        self.bot.create_line(
            bc, 0, bc, h, fill='#808080', width=1)
        self.bot.create_line(
            bc+1, 0, bc+1, h, fill='white', width=1)


class TableFrame:
    def __init__(self, graph, master, numkindrows, samplevar):
        self.graph = graph
        self.mod = graph.mod
        frame = self.frame = Frame(master, borderwidth=2, relief=GROOVE)
        row = 0
        self.marktime = StringVar()
        self.totsizevar = SizeVar()
        self.sampler = StringVar()
        self.sampler.set('R')

        fr = Frame(frame)  # For header
        om = OptionMenu(fr, self.sampler, 'R', 'L', 'R-L')
        om.grid(row=0, column=0, sticky=W)
        Label(fr, text='Sample').grid(row=0, column=1, sticky=W)
        Label(fr, textvariable=samplevar, background='black', foreground='yellow',
              ).grid(row=0, column=2, sticky=W, pady=3)
        Label(fr, text='at').grid(row=0, column=3, sticky=W)
        Label(fr, textvariable=self.marktime).grid(
            row=0, column=4, sticky=W)
        Label(fr, text='Total size = ').grid(
            row=1, column=0, columnspan=3, sticky=W)
        Label(fr, textvar=self.totsizevar).grid(
            row=1, column=3, columnspan=2, sticky=W)
        fr.grid(row=row, column=0, sticky=W)
        row += 1

        orow = row

        tb = Frame(frame)
        row = 0

        Label(tb, text="").grid(row=row, column=0)
        Label(tb, text="R", ).grid(row=row, column=1, sticky=E)
        Label(tb, text="%R").grid(row=row, column=2, sticky=E)
        Label(tb, text="R-L", ).grid(row=row, column=3, sticky=E)
        Label(tb, text="%L").grid(row=row, column=4, sticky=E)
        Label(tb, text="Kind").grid(row=row, column=5, sticky=W)

        row += 1
        self.profrows = []
        self.totrow = ProfileRow(tb, row)
        self.profrows.append(self.totrow)
        row += 1

        for i in range(numkindrows+1):
            profrow = ProfileRow(tb, row)
            self.profrows.append(profrow)
            row += 1

        row = orow
        tb.grid(row=row, column=0, sticky=W)

        # for next..
        row += 1

        self.totresize = 0
        self.kindwidth = ProfileRow.kindwidth

    def resize(self, dx, dy):
        dx = int(dx)
        self.totresize += dx
        charresize, extra = divmod(self.totresize, 7)
        newwidth = ProfileRow.kindwidth + charresize
        oldwidth = self.profrows[0].kind['width']
        if newwidth < 10:
            newwidth = 10
            dx = (newwidth - oldwidth) * 7 + extra
        for pr in self.profrows:
            pr.kind['width'] = newwidth
            pr.kindwidth = newwidth
            pr.kind['padx'] = extra / 2
            import textwrap
            kindtext = textwrap.fill(pr.kindtext, width=pr.kindwidth)
            pr.set_kind(pr.kindtext)

        return dx, dy

    def update(self, lsamp, rsamp):

        self.marktime.set(self.mod.time.asctime(
            self.mod.time.localtime(rsamp.stat.timemade)))

        return

        for pr in self.profrows:
            pr.clear()

        rdiv = float(rsamp.stat.size)
        ldiv = float(lsamp.stat.size)

        self.totrow.set_color_size_percent_kind(
            None,
            rsamp.stat.size,
            100.0,
            rsamp.stat.size - lsamp.stat.size,
            (rsamp.stat.size - lsamp.stat.size) * 100.0 / ldiv,
            '<Total>'
        )

        for i, r in enumerate(rsamp.rows):
            l = lsamp.kindrows[r.name]
            self.profrows[i+1].set_color_size_percent_kind(
                self.graph.getkindcolor(r.name),
                r.size,
                r.size * 100.0 / rdiv,
                r.size - l.size,
                (r.size - l.size) * 100.0 / ldiv,
                r.name)


class ColSpec:
    def __init__(self, tf, header, width, pos, render, idx=()):
        self.tf = tf
        self.header = header
        self.name = header
        self.width = width
        self.pos = pos
        self.render = render
        self.idx = idx

    def align(self, text):
        sp = ' '*(self.width - len(text))
        if self.pos == LEFT:
            text = text + sp
        elif self.pos == RIGHT:
            text = sp[:-1] + text + ' '
        else:
            assert 0
        assert len(text) == self.width
        return text


class TableFrame:
    def __init__(self, graph, master):
        self.graph = graph
        self.mod = graph.mod

        frame = self.frame = Frame(
            master,
            borderwidth=3,
            relief=SUNKEN
        )

        self.colspecs = {}
        self.colwidths = []

        def defcol(names, width, pos, put, idxfunc=lambda x: ()):
            if callable(put):
                put = [put]*len(names)
            self.colwidths.append(width)
            for name, put in zip(names, put):
                spec = ColSpec(self, name, width, pos, put, idxfunc(name))
                self.colspecs[name] = spec

        defcol(('A', 'B'), 2, LEFT, self.putcolor, lambda x: x)
        defcol(('Size', 'Count'), 7, RIGHT, [self.putsize, self.putcount])
        defcol(('%A:Tot', '%B:Tot'), 7, RIGHT,
               self.putpercent, lambda name: name[1])
        defcol(('B-A', 'A-B', 'Cumul'), 7, RIGHT, [self.putdiff, self.putdiff, self.putcumul],
               lambda name: [(), name.split('-')]['-' in name])
        defcol(('%A:Tot', '%B:Tot'), 7, RIGHT,
               self.putpercent, lambda name: name[1])
        defcol(('Kind',), 20, LEFT, self.putkind)

        width = 0
        for w in self.colwidths:
            width += w
        self.totxresize = 0
        self.totyresize = 0
        self.kindcol = self.colspecs['Kind']
        self.orgkindwidth = self.kindcol.width
        self.widthbeforekind = width - self.orgkindwidth
        self.minkindwidth = 10
        self.mintextheight = 2
        width += 1
        self.width = self.orgwidth = width

        wrap = NONE
        cursor = master['cursor']
        relief = FLAT
        self.minpadx = 3
        self.tothead = Text(
            frame,
            width=width,
            wrap=wrap,
            background='#ccc',
            height=2,
            padx=self.minpadx,
            relief=relief,
            cursor=cursor,
        )

        self.rowhead = Text(
            frame,
            width=width,
            wrap=wrap,
            background='#ccc',
            height=1,
            padx=self.minpadx,
            relief=relief,
            cursor=cursor,
        )

        self.tsframe = Frame(frame)

        self.textminpady = 2
        self.text = Text(
            self.tsframe,
            width=width,
            wrap=wrap,
            height=21,
            background='#e0e0e0',
            relief=relief,
            takefocus=0,
            cursor=cursor,
            padx=self.minpadx,
            pady=self.textminpady,
        )

        self.scrollbar = Scrollbar(
            self.tsframe,
            width=10,
            orient=VERTICAL,
            command=self.text.yview
        )
        self.scrollbar_totwidth = int(
            self.scrollbar['width']) + 6  # width + padding

        self.uses_scrollbar = 0
        self.auto_scrollbar = 1

        self.orgtextheight = int(self.text['height'])

        padx = 0
        pady = 0
        self.tothead.pack(anchor=N+W, padx=padx, pady=pady)
        self.rowhead.pack(anchor=N+W, padx=padx, pady=pady)
        self.text.pack(side=LEFT, anchor=N+W, padx=padx, pady=pady)
        self.tsframe.pack(anchor=N+W, padx=padx, pady=pady)

    def setchdim(self):
        self.text.update()
        self.chdx = float(self.text.winfo_width()) / self.width
        self.chdy = float(self.text.winfo_height()) / self.orgtextheight
        self.chdx = int(round(self.chdx))
        self.chdy = int(round(self.chdy))
        self.pixwidth = self.width * self.chdx
        self.pixheight = self.width * self.chdy

    def putcolor(self, col):
        if self.colorow.name == '<Total>':
            text = col.align(' ')
            color = '#e0e0e0'
        else:
            color = self.graph.getkindcolor(self.colorow.name),
            text = col.align('@')
        self.text.insert('end', text, (color,))
        self.text.tag_config(color, foreground=color, background='#e0e0e0',
                             font=('terminal', '12', 'bold'),)

    def putcount(self, col):
        self.valmode = 'Count'
        count = self.colorow.count
        self.cumulval += count
        self.putval(col, count)

    def putsize(self, col):
        self.valmode = 'Size'
        size = self.colorow.size
        self.cumulval += size
        self.putval(col, size)

    def putval(self, col, val):
        self.curval = val
        self.ap(col.align(sizestring(val)))

    def putpercent(self, col):
        a = self.statbyname[col.idx]

        if self.valmode == 'Count':
            ref = a.count
        elif self.valmode == 'Size':
            ref = a.size

        if ref:
            ps = percentstring(self.curval * 100.0 / ref)
        else:
            ps = '---'
        self.ap(col.align(ps))

    def putdiff(self, col):

        a, b = self.rowbyname[col.idx[0]], self.rowbyname[col.idx[1]]
        if self.valmode == 'Count':
            a, b = a.count, b.count
        elif self.valmode == 'Size':
            a, b = a.size, b.size

        self.putval(col, a - b)

    def putcumul(self, col):
        self.putval(col, self.cumulval)

    def putkind(self, col):
        # Must be last!
        import textwrap
        wraplines = textwrap.wrap(self.colorow.name, width=col.width)
        self.ap(col.align(wraplines[0]))
        if len(wraplines) > 1:
            initial = '\n'+' '*(self.widthbeforekind)
            for line in wraplines[1:]:
                self.ap(initial+col.align(line))

    def setmode(self, mode, numkindrows):
        self.mode = mode
        self.numkindrows = numkindrows
        self.mcontrols = self.graph.mcontrolbyname
        self.stats = self.graph.stats
        self.cols = [self.colspecs[x.strip()]
                     for x in mode.split(' ') if x.strip()]
        self.controlnames = {}
        name = self.cols[0].idx
        self.colorcontrol = self.mcontrols[name]
        self.controlnames[name] = 1
        self.controls = [self.colorcontrol]
        self.lastidxs = [None]
        for i, co in enumerate(self.cols):
            idx = co.idx
            if not isinstance(idx, (tuple, list)):
                idx = (idx,)
            for idx in idx:
                if idx not in self.controlnames:
                    self.controls.append(self.mcontrols[idx])
                    self.controlnames[idx] = 1
                    self.lastidxs.append(None)

    def setscrollbar(self, sb):
        if sb == self.uses_scrollbar:
            return
        self.uses_scrollbar = sb
        w = self.scrollbar_totwidth
        if sb:
            self.resize(-w, 0, setscrollbar=0)
            self.scrollbar.pack(side=LEFT, fill=Y)
            self.text['yscrollcommand'] = self.scrollbar.set
        else:
            self.resize(w, 0, setscrollbar=0)
            self.scrollbar.pack_forget()
            self.text['yscrollcommand'] = None

    def update_simple(self, lsamp, rsamp):
        t = self.text
        t.delete('1.0', '100.0')
        t.insert('1.0', str(rsamp.stat))

    def update(self, force=0, setscrollbar=1):

        stats = self.stats

        idxs = [max(0, min(control.sample, len(stats)-1))
                for control in self.controls]
        if (idxs == self.lastidxs) and not force:
            return

        self.lastidxs = idxs

        self.text['state'] = self.tothead['state'] = self.rowhead['state'] = NORMAL

        self.text.delete('1.0', END)
        self.tothead.delete('1.0', END)
        self.rowhead.delete('1.0', END)

        if not stats:
            self.tothead.insert('end', '-- No Sample --')
            self.text['state'] = self.tothead['state'] = self.rowhead['state'] = DISABLED
            return

        self.statbyname = {}
        statbyidx = []
        for i, control in enumerate(self.controls):
            stat = stats[idxs[i]]
            statbyidx.append(stat)
            self.statbyname[control.name] = stat

        samps = self.samps = [
            Sample(self.mod, statbyidx[0], self.controls[0].marker.name, idxs[0],
                   numkindrows=self.numkindrows,
                   statype=self.graph.display.statype
                   )]

        self.colorsamp = samps[0]
        if len(self.controls) > 1:
            samps.append(Sample(self.mod, statbyidx[1], self.controls[1].marker.name, idxs[1],
                                relative=samps[0]))

            self.relsamp = samps[1]

        t = self.tothead

        n = max([len(str(samp.index)) for samp in samps])
        for samp in samps:
            t.insert('end', 'Sample %s: ' % samp.name)
            t.insert('end', ('%%%dd' % n) % samp.index, ('index',))
            t.insert('end', ' at %s\n' % (samp.datetime))

        t.tag_configure('index', background='#e0e0e0')

        t = self.rowhead

        self.sizes = [float(samp.stat.size) for samp in samps]

        for col in self.cols:
            t.insert('end', col.align(col.header), ('header',))
        t.insert('end', '\n')

        t = self.text

        self.ap = lambda text: t.insert('end', text)

        self.colorow = Row(samps[0].count, samps[0].size, '<Total>')
        self.rowbyname = self.statbyname
        self.cumulval = 0
        for col in self.cols:
            col.render(col)
        self.ap('\n\n')

        self.cumulval = 0
        for i, a in enumerate(samps[0].rows):
            self.colorow = a
            if len(samps) > 1:
                self.rowbyname = {
                    samps[0].name: a,
                    samps[1].name: samps[1].kindrows[a.name]
                }
            for col in self.cols:
                col.render(col)
            self.ap('\n')

        if setscrollbar and self.auto_scrollbar:
            numrows = int(self.text.index('end').split('.')[0])-2
            h = int(self.text['height'])
            needs_scrollbar = numrows > h
            if needs_scrollbar != self.uses_scrollbar:
                self.setscrollbar(needs_scrollbar)

        self.text['state'] = self.tothead['state'] = self.rowhead['state'] = DISABLED

    def resize(self, dx, dy, setscrollbar=1):
        dx = int(dx)
        oldwidth = self.pixwidth
        newwidth = self.pixwidth + dx
        if newwidth < self.chdx * 2:
            newwidth = self.chdx * 2

        self.pixwidth = newwidth
        dx = newwidth - oldwidth

        charwidth, extra = divmod(newwidth, self.chdx)

        self.kindcol.width = max(
            charwidth - self.widthbeforekind - 1, self.minkindwidth)

        self.totxresize += dx
        for t in (self.tothead, self.rowhead, self.text):
            t['width'] = charwidth
            t['padx'] = self.minpadx + extra / 2

        dy = int(dy)

        rowresize, extra = divmod(self.totyresize + dy, self.chdy)
        newheight = self.orgtextheight + rowresize
        oldheight = int(self.text['height'])
        if newheight < self.mintextheight:
            newheight = self.mintextheight
            dy = (newheight - oldheight) * self.chdy + extra
        self.totyresize += dy

        self.text['height'] = newheight
        self.text['pady'] = self.textminpady + extra / 2

        self.update(force=1, setscrollbar=1)

        return dx, dy


class Filler:
    def __init__(self, master):
        self.frame = self.can = Canvas(
            master,
            # background='blue',
            width=0,
            height=0)

    def getsize(self):
        return int(self.can['width']), int(self.can['height']),

    def setsize(self, w, h):
        self.can.configure(
            width=w,
            height=h
        )

    def resize(self, dw, dh):
        w, h = self.getsize()
        self.setsize(max(0, w + dw), max(0, h + dh))


class Row:
    def __init__(self, count, size, name):
        self.count = count
        self.size = size
        self.name = name


class Sample:
    def __init__(self, mod, stat, name, index, numkindrows=None, statype='Size', relative=None):
        self.stat = stat
        self.size = stat.size
        self.count = stat.count
        self.name = name
        self.index = index

        self.datetime = mod.time.asctime(mod.time.localtime(stat.timemade))

        self.kindrows = {}

        if numkindrows is not None:
            rows = stat.get_rows_n_and_other(numkindrows, statype)
            for r in rows:
                self.kindrows[r.name] = r
        else:
            kinds = []
            oidx = None
            for row in relative.rows:
                if row.name == '<Other>':
                    oidx = len(kinds)
                    continue
                else:
                    kinds.append(row.name)

            rows = stat.get_rows_of_kinds(kinds)
            size = 0
            count = 0
            for i, row in enumerate(rows):
                kind = kinds[i]
                if row is None:
                    row = Row(0, 0, kind)
                self.kindrows[kind] = row
                size += row.size
                count += row.count

            if oidx is not None:
                other = Row(stat.count - count, stat.size - size, '<Other>')
                rows[oidx:oidx] = [other]
                self.kindrows['<Other>'] = other

        self.rows = rows


class ProfileBrowser:
    colors = ("red", "green", "blue", "yellow", "magenta", "cyan", 'white')
    numkindrows = 10

    def __init__(self, app, filename):
        self.inited = 0
        self.app = app
        self.mod = mod = app.mod
        self.master = master = app.root
        if filename:
            filename = mod.path.abspath(filename)
            self.initialdir = mod.path.dirname(filename)
        else:
            self.initialdir = mod.os.getcwd()

        self.frame = frame = Toplevel(
            master,
            # background='#bbb'
        )
        #frame['cursor'] = 'umbrella'
        # frame.resizable(True,True)

        self.menubar = Frame(self.frame, relief=RAISED, bd=2)

        self.filebutton = Menubutton(self.menubar, text='File')
        self.filemenu = Menu(self.filebutton)
        self.filebutton['menu'] = self.filemenu
        self.filemenu.add_command(
            label='New Profile Browser', command=self.cmd_new)
        self.filemenu.add_command(label='Open Profile', command=self.cmd_open)
        self.filemenu.add_command(label='Close Window', command=self.cmd_close)
        self.filemenu.add_command(
            label='Clear Cache', command=self.cmd_clear_cache)
        self.filemenu.add_command(label='Exit', command=self.cmd_exit)
        self.panebutton = Menubutton(self.menubar, text='Pane')
        self.panemenu = Menu(self.panebutton)
        self.panebutton['menu'] = self.panemenu

        choices = [
            ('Bars', 'Lines'),
            ('Size', 'Count'),
        ]

        self.graphtypevar = StringVar()
        self.graphbutton = self.modechooser(
            self.menubar, 'Graph', choices,
            self.graphtypevar, self.cmd_graphtype)

        choices = [
            ('A', 'B'),
            ('Size', 'Count'),
            ('%A:Tot', '%B:Tot'),
            ('Cumul', 'A-B', 'B-A'),
            ('%A:Tot', '%B:Tot'),
            ('Kind',),
        ]

        self.var_tablemode = StringVar()
        self.tablebutton = Menubutton(self.menubar, text='Table')
        self.tablemenu = Menu(self.tablebutton)
        self.tablebutton['menu'] = self.tablemenu
        self.headermenu = Menu(self.tablebutton, title='Table header')
        self.addmodechooser(
            self.headermenu,
            choices,
            self.var_tablemode,
            self.cmd_tablemode
        )
        self.tablemenu.add_cascade(label='Header', menu=self.headermenu)
        self.var_tablescrollbar = StringVar()
        self.tablescrollbarmenu = Menu(
            self.tablebutton, title='Table scrollbar')

        self.addmodechooser(
            self.tablescrollbarmenu,
            [('Auto', 'On', 'Off')],
            self.var_tablescrollbar,
            self.cmd_tablescrollbar
        )

        self.tablemenu.add_cascade(
            label='Scrollbar',
            menu=self.tablescrollbarmenu)

        self.windowmenu = WindowMenu(self.menubar, self.app.var_window)
        self.window = app.add_window_frame(self.frame, self.windowmenu)

        self.helpbutton = Menubutton(self.menubar, text='Help')
        self.helpmenu = Menu(self.helpbutton)
        self.helpbutton['menu'] = self.helpmenu
        self.helpmenu.add_command(label='About', command=self.cmd_about)
        self.helpmenu.add_command(label='Help', command=self.cmd_help)

        self.ctrlframe = Frame(
            self.frame,
            bd=2,
            relief=GROOVE,
            # background='#999',


        )

        self.exitbutton = Button(self.ctrlframe, text='Exit', command=self.cmd_exit,
                                 background='red')

        self.set_filename(filename)

        self.id_collect = None
        self.collecting = IntVar()
        self.collecting.set(0)
        self.collectbutton = Checkbutton(self.ctrlframe, text='Collect',
                                         variable=self.collecting,
                                         command=self.cmd_collect,
                                         relief=RAISED)

        self.stats = Stats(self.mod)

        self.disptab = Frame(self.frame,
                             # relief=SUNKEN,
                             # bd=3
                             )

        self.display = Display(self.disptab,
                               scale_table=AxisControl.scale_table,
                               numkindrows=self.numkindrows,
                               getkindcolor=self.getkindcolor,
                               )

        self.xcontrol = AxisControl(self.ctrlframe,
                                    name='X',
                                    range=self.display.xrange,
                                    grid=self.display.xgrid,
                                    unit='samples',
                                    rangecommand=self.display.setxrange,
                                    gridcommand=self.display.setxgrid
                                    )

        self.ycontrol = AxisControl(self.ctrlframe,
                                    name='Y',
                                    range=self.display.yrange,
                                    grid=self.display.ygrid,
                                    unit='bytes',
                                    rangecommand=self.display.setyrange,
                                    gridcommand=self.display.setygrid,
                                    autocommand=self.display.cmd_yrange_auto
                                    )

        self.display.xcontrol = self.xcontrol
        self.display.ycontrol = self.ycontrol

        self.mcontrols = []
        self.mcontrolbyname = {}
        for name in ('A', 'B'):
            marker = self.display.new_xmarker(name)
            control = MarkerControl(
                self.ctrlframe, marker, self.update_tableframe)
            marker.set_poscommand(control.setsample)
            self.mcontrols.append(control)
            self.mcontrolbyname[name] = control

        self.var_showcontrol = BooleanVar()
        self.var_showcontrol.set(1)
        self.panemenu.add_checkbutton(
            label='Show Control Panel',
            variable=self.var_showcontrol,
            command=self.cmd_showcontrol)

        self.var_showgraph = BooleanVar()
        self.var_showgraph.set(1)
        self.panemenu.add_checkbutton(
            label='Show Graph',
            variable=self.var_showgraph,
            command=self.cmd_showgraph)

        self.var_showtable = BooleanVar()
        self.var_showtable.set(1)
        self.panemenu.add_checkbutton(
            label='Show Table',
            variable=self.var_showtable,
            command=self.cmd_showtable)

        tf = self.tf = TableFrame(self, self.disptab)
        d_t = self.d_t = PaneDiv(self.disptab, movecommand=self.cmd_dt_moved)

        self.xcontrol.frame.grid(row=0, column=0, padx=3, pady=3, sticky=W)
        self.ycontrol.frame.grid(row=1, column=0, padx=3, pady=3)
        self.mcontrols[0].frame.grid(
            row=0, column=1, columnspan=1, sticky=W, padx=3, pady=3)
        self.mcontrols[1].frame.grid(
            row=1, column=1, columnspan=1, sticky=W, padx=3, pady=3)
        self.exitbutton.grid(row=0, column=2, padx=3, pady=3)
        self.collectbutton.grid(row=0, column=3, padx=3, pady=3)

        self.filler = Filler(self.frame)

        self.filebutton.pack(side=LEFT)
        self.panebutton.pack(side=LEFT)
        self.graphbutton.pack(side=LEFT)
        self.tablebutton.pack(side=LEFT)
        self.windowmenu.button.pack(side=LEFT)
        self.helpbutton.pack(side=LEFT)

        self.menubar.grid(column=0, columnspan=4, sticky=N+W+E)
        self.gridmain()

        frame.bind('<Map>', self.event_map)

        self.tf.setmode(self.var_tablemode.get(), self.numkindrows)

        self.load_filename(filename)

        d_t.frame.update_idletasks()
        d_t.setheight(max(self.display.frame.winfo_height(),
                          tf.frame.winfo_height()))

        d_t.frame.update_idletasks()
        self.minsize = (500, 400)
        self.maxsize = (self.frame.winfo_screenwidth(),
                        self.frame.winfo_screenheight())
        minsizes = {
            # (ctrl, disp, tab) : (width, height)
            (0, 0, 0): (270, 25),
            (1, 0, 0): (363, 61),
            (0, 1, 0): (270, 131),
            (1, 1, 0): (270, 131),
        }

        self.setusergeometry()

        def initfinal():
            self.tf.setchdim()

            rx = self.frame.winfo_rootx() + self.frame.winfo_width()
            self.tf_wanted_margin = rx - \
                (self.tf.frame.winfo_rootx() + self.tf.frame.winfo_width())

            self.lastw = self.frame.winfo_width()
            self.lasth = self.frame.winfo_height()
            self.in_configure = 0
            frame.bind('<Configure>', self.event_configure)
            self.inited = 1

        initfinal()
        # self.frame.after_idle(initfinal)

    def cmd_about(self):
        self.cmd_help('about')

    def cmd_help(self, pickname='help'):
        os = self.mod.os
        ocursor = self.frame.winfo_toplevel()['cursor']
        try:
            self.frame.winfo_toplevel()['cursor'] = 'watch'
            self.frame.update()
            m = self.mod.Text.gsltextviewer(
                self.frame,
                inpickle=getattr(self.mod.pbhelp, pickname)
                # htmloutfile='/tmp/x.html',
            )

            self.app.add_window_frame(m)
        finally:
            self.frame.winfo_toplevel()['cursor'] = ocursor

    def cmd_clear_cache(self):
        self.stats.clear_cache()

    def cmd_close(self):
        self.frame.destroy()

    def cmd_collect(self, *args):
        # self.afterfunc()
        # self.frame.after(1, self.afterfunc) # Turn on button first.??

        if self.collecting.get():
            self.event_collect()
        else:
            if self.id_collect is not None:
                self.frame.after_cancel(self.id_collect)
                self.id_collect = None

    def event_collect(self):
        o, n = self.stats.collect()
        if n:
            if o != self.display.numstats:
                self.display.load_stats(self.stats)
            else:
                st = self.stats[-n:]
                self.display.add_stats(st)
            for c in self.mcontrols:
                c.setnumsamples(len(self.stats))

        self.id_collect = self.frame.after(1000, self.event_collect)

    def cmd_dt_moved(self, dx):
        # The division between display and table panes moved.

        # Disable configure event handling while we are resizing.
        self.in_configure += 1

        # Right x position of enclosing frame

        rx = self.frame.winfo_rootx() + self.frame.winfo_width()

        # Right margin between pane divider and enclosing window

        mx = rx - (self.d_t.frame.winfo_rootx() + self.d_t.frame.winfo_width())

        # Don't move pane divider outside window
        dx = min(dx, mx)

        # Right margin between table and enclosing window
        # before resizing
        mx = rx - (self.tf.frame.winfo_rootx() + self.tf.frame.winfo_width())

        dx, _ = self.display.resize(dx, 0)

        wanted_margin = self.tf_wanted_margin

        # After move
        mx -= dx
        self.tf.resize(mx - wanted_margin, 0)

        self.display.moveback()

        self.in_configure -= 1

    def cmd_exit(self):
        self.app.exit()

    def cmd_graphtype(self):
        self.display.setgraphtype(self.graphtypevar.get(), self.stats)
        self.cmd_tablemode()

    def cmd_new(self):
        self.app.new_profile_browser(self.filename)

    def cmd_open(self):
        op = tkinter.filedialog.Open(self.frame,
                                     # ? Should we have default extension or not??
                                     # defaultextension='.hpy',
                                     initialdir=self.initialdir,
                                     filetypes=[('Heapy data files', '.hpy'),
                                                ('All files', '*')
                                                ]
                                     )
        filename = op.show()
        if filename:
            self.load_filename(filename)

    def cmd_showcontrol(self):
        self.grid_things()

    def cmd_showgraph(self):
        if self.var_showgraph.get() and self.var_showtable.get():
            self.tf.resize(-self.tf.totxresize, 0)
            self.display.resize(self.display.orgwidth - self.display.botx, 0)
            self.display.moveback()
        self.grid_things()

    cmd_showtable = cmd_showgraph

    def cmd_tablemode(self):
        self.tf.setmode(self.var_tablemode.get(), self.numkindrows)
        self.tf.update()

    def cmd_tablescrollbar(self):
        tf = self.tf
        s = self.var_tablescrollbar.get()
        if s == 'Auto':
            tf.auto_scrollbar = 1
            tf.update(force=1, setscrollbar=1)
        elif s == 'On':
            tf.auto_scrollbar = 0
            tf.setscrollbar(1)
        elif s == 'Off':
            tf.auto_scrollbar = 0
            tf.setscrollbar(0)
        else:
            assert 0

    def setusergeometry(self):
        # Make the geometry of the window be user-specified
        # This is called after Tk has determined the size
        # of the window needed for the initial widget configuration.
        # The size is not to be changed after that, other than
        # on user request.
        # I couldn't just do frame.geometry(frame.geometry()) because,
        # presumably, of a bug in the Tk and/or wm I am using. I hope
        # this works for all systems .. Notes  26 Oct 2005.

        self.frame.update()
        g = '%dx%d+%d+%d' % (
            self.frame.winfo_width(),
            self.frame.winfo_height(),
            self.frame.winfo_rootx(),
            self.frame.winfo_rooty())
        self.frame.geometry(g)

    def modechooser(self, frame, name, choices, cmdvar, command):

        button = Menubutton(frame, text=name)
        menu = Menu(button)
        button['menu'] = menu

        self.addmodechooser(menu, choices, cmdvar, command)
        return button

    def addmodechooser(self, menu, choices, cmdvar, command):

        def setcmdvar():
            cmdvar.set(' '.join([v.get() for v in vars]))

        def cmd():
            setcmdvar()
            command()

        vars = []
        for ch in choices:
            var = StringVar()
            vars.append(var)
            var.set(ch[0])
            for a in ch:
                menu.add_radiobutton(
                    command=cmd,
                    label=a,
                    value=a,
                    variable=var,
                    #font=('Courier','12', 'bold'),
                    #font=('Helvetica','12', 'bold'),
                    columnbreak=(a == ch[0])
                )

        setcmdvar()

    def grid_things(self):
        ow = self.frame.winfo_width()
        oh = self.frame.winfo_height()

        self.ctrlframe.grid_forget()
        self.display.frame.grid_forget()
        self.d_t.frame.grid_forget()
        self.tf.frame.grid_forget()
        self.disptab.grid_forget()
        self.filler.frame.grid_forget()

        self.gridmain()

        self.frame.update_idletasks()
        self.sizewidgets()

    def gridmain(self):

        row = 1

        c = self.var_showcontrol.get()
        if c:
            self.ctrlframe.grid(row=row, column=0,
                                columnspan=3, padx=3, pady=3, sticky=W)
            row += 1

        column = 0

        g = self.var_showgraph.get()
        t = self.var_showtable.get()
        gt = (g, t)
        if g:
            self.display.frame.grid(row=0, column=column, sticky=N+W,
                                    padx=3, pady=3
                                    )
            column += 1

        if g and t:
            self.d_t.frame.grid(row=0, column=column, sticky=N+W)
            column += 1
        if t:
            self.tf.frame.grid(row=0, column=column, sticky=N+W, padx=3, pady=3
                               )
        if g or t:
            self.disptab.grid(row=row, column=0,
                              sticky=N+W,
                              # padx=3,pady=3,
                              )
            row += 1
        self.filler.setsize(0, 0)
        self.filler.frame.grid(row=row, column=3, sticky=N+W)

        self.frame.resizable(1, 1)

    def event_configure(self, event):
        if event.widget is not self.frame:
            return

        if not self.inited:
            return

        if self.in_configure:
            return

        curw = self.frame.winfo_width()
        curh = self.frame.winfo_height()
        if curw == self.lastw and curh == self.lasth:
            return

        self.in_configure += 1

        self.lastw = curw
        self.lasth = curh

        self.sizewidgets()

        self.in_configure -= 1

    def sizewidgets(self):
        self.frame.update()
        curw = self.frame.winfo_width()
        curh = self.frame.winfo_height()

        mbx = self.menubar.winfo_rootx()
        mby = self.menubar.winfo_rooty()

        sfs = []
        if self.var_showgraph.get():
            sfs.append(self.display)
        if self.var_showtable.get():
            sfs.append(self.tf)

        if not sfs:
            sfs.append(self.filler)

        dys = {}
        didh = 0
        for sf in sfs:
            f = sf.frame
            diy = f.winfo_rooty()
            dih = f.winfo_height()

            ch = diy - mby + dih
            dy = curh - ch - 7
            didh = didh or dy
            dys[sf] = dy

        if self.var_showtable.get():
            f = self.tf.frame
        elif self.var_showgraph.get():
            f = self.display.frame
        else:
            f = self.filler.frame

        fx = f.winfo_rootx()
        fw = f.winfo_width()

        cw = fx - mbx + fw

        fdw = curw - cw - 6

        if f is self.filler.frame and not self.var_showcontrol.get():
            fdw = curw - self.filler.getsize()[0] - 3

        if didh or fdw:
            if self.var_showgraph.get() and self.var_showtable.get():
                dprop = float(self.display.frame.winfo_width())
                dprop = dprop / (dprop + self.tf.frame.winfo_width())
                dx, dy = self.display.resize(fdw * dprop, dys[self.display])
                self.tf.resize(fdw - dx, dys[self.tf])
                self.frame.update_idletasks()
                self.d_t.setheight(max(self.display.frame.winfo_height(),
                                       self.tf.frame.winfo_height()))

            elif self.var_showgraph.get():
                self.display.resize(fdw, dys[self.display])
            elif self.var_showtable.get():
                self.tf.resize(fdw, dys[self.tf])
            else:
                self.filler.resize(fdw, dys[self.filler])
                self.filler.setsize(self.filler.getsize()[0], 1000)
            if self.var_showgraph.get():
                self.display.moveback()

        #self.resize(dw, dh)

    def resize(self, dw, dh):
        self.display.resize(dw, dh)
        # self.frame.wm_geometry('')

    def event_map(self, event):
        self.frame.unbind('<Map>')
        self.frame.bind('<Unmap>', self.event_unmap)
        self.frame.lift()

    def event_unmap(self, event):
        self.frame.unbind('<Unmap>')
        self.frame.bind('<Map>', self.event_map)

    def load_filename(self, filename):
        ocursor = self.frame.winfo_toplevel()['cursor']
        try:
            self.frame.winfo_toplevel()['cursor'] = 'watch'
            self.frame.update()
            if filename:
                filename = self.mod.path.abspath(filename)
            try:
                self.stats.open(filename)
            except Exception:
                __import__('traceback').print_exc()
                etype, value, tb = self.mod._root.sys.exc_info()
                tkinter.messagebox.showerror(
                    master=self.frame,
                    message=(
                        "Error when loading\n%r:\n" % filename +
                        "%s" % ''.join(self.mod._root.traceback.format_exception_only(
                            etype, value)))
                )
            else:
                self.display.load_stats(self.stats)

                for c in self.mcontrols:
                    c.setnumsamples(len(self.stats))
                # self.scontrol.trackcommand(1)
                self.set_filename(filename)

                self.xrange_fit()
                self.display.xview_moveto(0)
                self.mcontrols[1].settracking(0)
                self.mcontrols[0].settracking(1)
                self.yrange_fit()
                self.tf.update(force=1)
                if filename:
                    self.initialdir = self.mod.path.dirname(filename)

        finally:
            self.frame.winfo_toplevel()['cursor'] = ocursor

    def update_tableframe(self):
        self.tf.update()

    def getkindcolor(self, kind):
        if kind == '<Other>':
            return 'black'
        else:
            return self.colors[abs(hash(kind)) % len(self.colors)]

    def set_filename(self, filename):
        self.filename = filename
        if not filename:
            filename = '<No File>'
        title = 'Heapy Profile Browser: %s' % filename
        self.window.title(title)

    def setnormpos(self):
        self.setscrollregion()
        if self.ymax >= self.yrange:
            self.yrange_fit()
        if self.xi0 is None:
            self.drawxaxis()
        else:
            self.updatexaxis()

        self.track()

    def redraw_all(self):
        pass

    def trackoff(self):
        self.rcontrol.settracking(0)

    def xrange_fit(self):
        self.xcontrol.fit(len(self.stats))

    def yrange_fit(self):
        self.display.yrange_auto(force=1)


class _GLUECLAMP_:
    _imports_ = (
        '_parent:Use',
        '_parent:pbhelp',
        '_root.guppy.etc:textView',
        '_root:hashlib',
        '_root:os',
        '_root.os:path',
        '_root:time',
        '_root.guppy.gsl:Text',
    )

    def pb(self, filename=None):
        """pb( [filename: profilefilename+])

Create a Profile Browser window.

Argument
    filename: profilefilename+
        The name of a file containing profile data.
See also
    Heapy Profile Browser[1]
    Screenshot[2]
References
    [0] heapy_Use.html#heapykinds.Use.pb
    [1] ProfileBrowser.html
    [2] pbscreen.jpg"""

        pa = ProfileApp(self)
        pa.new_profile_browser(filename)
        pa.mainloop()

    def tpg(self):
        self('/tmp/x.hpy')



import os.path
import re
import sys
from setuptools import setup, Extension
from distutils.command.install import INSTALL_SCHEMES

for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']


setsc = Extension("guppy.sets.setsc", [
                      "src/sets/sets.c",
                      "src/sets/bitset.c",
                      "src/sets/nodeset.c"
                  ])

heapyc = Extension("guppy.heapy.heapyc", [
                       'src/heapy/heapyc.c',
                       'src/heapy/stdtypes.c'
                   ])


def doit():
    if sys.version_info.major < 3:
        print('''\
setup.py: Error: This guppy package only supports Python 3.
You can find the original Python 2 version, Guppy-PE, here:
http://guppy-pe.sourceforge.net/''', file=sys.stderr)
        sys.exit(1)
    if sys.implementation.name != 'cpython':
        print('''\
setup.py: Warning: This guppy package only supports CPython.
Compilation failure expected, but continuting anyways...''', file=sys.stderr)

    with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
        long_description = f.read()

    with open('guppy/_version.py', 'r') as versionfile:
        version = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]$',
                            versionfile.read(), re.M)
        version = version.group(1)

    setup(name="guppy3",
          version=version,
          description="Guppy 3 -- Guppy-PE ported to Python 3",
          long_description=long_description,
          long_description_content_type='text/markdown',
          author="YiFei Zhu",
          author_email="zhuyifei1999@gmail.com",
          url="https://github.com/zhuyifei1999/guppy3/",
          license='MIT',
          packages=[
              "guppy",
              "guppy.etc",
              "guppy.gsl",
              "guppy.heapy",
              "guppy.heapy.test",
              "guppy.sets",
          ],
          ext_modules=[setsc, heapyc],
          python_requires='>=3.6',
          classifiers=[
              "Programming Language :: Python :: 3",
              "Programming Language :: Python :: Implementation :: CPython",
              "Programming Language :: C",
              "License :: OSI Approved :: MIT License",
              "Operating System :: OS Independent",
              "Development Status :: 4 - Beta",
              "Topic :: Software Development :: Debuggers",
              "Environment :: Console",
              "Intended Audience :: Developers",
          ])


doit()



import code


class Console(code.InteractiveConsole):
    EOF_key_sequence = '-'

    def __init__(self, stdin, stdout, locals=None, filename="<console>"):
        self.stdin = stdin
        self.stdout = stdout
        code.InteractiveConsole.__init__(self, locals, filename)

    def raw_input(self, prompt=""):
        """Write a prompt and read a line.

        The returned line does not include the trailing newline.
        When the user enters the EOF key sequence, EOFError is raised.

        """
        self.write(prompt)
        line = self.stdin.readline()
        if not line:
            raise EOFError
        line = line.rstrip()
        if line == self.EOF_key_sequence:
            raise EOFError
        else:
            return line

    def write(self, data):
        self.stdout.write(data)
        self.stdout.flush()



def str_as_atr(s):
    if s == '_':
        return []
    atr = []
    i = 0
    while i < len(s):
        v = s[i]
        if i + 1 < len(s) and s[i+1].isdigit():
            n = 0
            i = i + 1
            while i < len(s) and s[i].isdigit():
                n = n * 10 + int(s[i])
                i += 1
        else:
            i += 1
            n = 1
        for j in range(n):
            atr.append(v)
    return atr


def atr_as_str(atr):
    if not atr:
        return '_'
    prefl = []
    prestr = ''.join([str(x) for x in atr])
    i = 0
    while i < len(prestr):
        c = prestr[i]
        j = i+1
        while j < len(prestr) and prestr[j] == c:
            j += 1
        if j - i > 2:
            prefl.append(c)
            prefl.append(str(j-i))
        else:
            while i < j:
                prefl.append(c)
                i += 1
        i = j
    return ''.join(prefl)


def str_as_ixl(s):
    return [ord(ch)-ord('a') for ch in str_as_atr(s)]


def ixl_as_str(ixl):
    return atr_as_str([chr(ix + ord('a')) for ix in ixl])


class Paths:
    def __init__(self, mod, rp, key, extended=True, andsets=(), variant=2):
        self.mod = mod
        self._hiding_tag_ = mod._hiding_tag_
        self.key = key
        self.rp = rp
        self.extended = extended
        self.srcrow = rp.get_row(self.key)
        self.variant = variant
        self.andsetbyname = {}
        row = self.srcrow
        while row is not None:
            self.andsetbyname[row.ixlstr] = mod.Use.Anything
            row = row.parent

        if isinstance(andsets, dict):
            self.andsetbyname.update(andsets)
        elif isinstance(andsets, (tuple, list)):
            row = self.srcrow
            for i, s in enumerate(andsets):
                if row is None:
                    raise ValueError('andsets argument is too long')
                if s is not None:
                    self.andsetbyname[row.ixlstr] = s
                row = row.parent
        else:
            raise TypeError('andsets argument must be dict, tuple, or list')

        mod.OutputHandling.setup_printing(
            self,
            stop_only_when_told=variant >= 2)

    def get_str_of_path_component_singleton(self, set):
        return set.brief.lstrip('<1 ').rstrip('>')

    def source_to_target_info(self):
        src = 'Source'
        tgt = 'Target'
        via = 'Via'

        row = self.srcrow
        indent = 0
        while row is not None:
            if row.parent is None:
                a = tgt
            elif row is srcrow:
                a = src
            else:
                a = via
            a = ' '*indent + a
            name = row.ixlstr
            a = a + ' ' + ' '*(8+srcrow.depth*indinc -
                               len(name)-len(a)) + name + ': '
            yield a+row.getsummary(mod.line_length-len(a))
            row = row.parent
            indent += indinc

    def _oh_get_label(self):
        return 'Paths from source %r to target %r.' % (self.srcrow.ixlstr, '_')

    def _oh_get_line_iter(self):
        return getattr(self, 'get_line_iter_%s' % (self.variant,))()

    def _oh_get_more_state_msg(self, startindex, lastindex):
        return ''

    def get_line_iter_1(self):
        # Original variant indenting from left to right

        mod = self.mod

        srcrow = self.srcrow
        srcset = srcrow.set

        indinc = 2
        if srcrow.depth >= 10:
            indinc = 1

        def genlines(row, ks, indent=0):
            par = row.parent
            for key, i, set in ks:
                sidx = '%s[%d]' % (row.ixlstr, i)

                if self.extended:
                    strsing = self.get_str_of_path_component_singleton(set)
                else:
                    strsing = ''
                vline = '%s %s %s %s' % (
                    key,
                    ' '*(40-len(key) - len(sidx)),
                    sidx,
                    strsing
                )

                yield vline

                if par is None:
                    continue

                def get_nks(key, set):
                    parset = set.referents & par.set
                    for i, p in enumerate(parset.byid.parts):
                        rels = mod.Path.relations(set.theone, p.theone)
                        for rel in rels:
                            if rel is mod.Path.identity:
                                continue
                            if rel is mod.Path.norelation:
                                k = '??'
                            else:
                                k = str(rel) % ''
                            k = ' '*(indent+indinc)+k
                            yield k, i, p

                for line in genlines(par, get_nks(key, set), indent+indinc):
                    yield line

        def get_ks():
            for i, s in enumerate(srcset.byid.parts):
                k = '[%d]  ' % i
                k = k + (' -'*20)[:36-len(k)-srcrow.depth]
                yield k, i, s

        for line in genlines(srcrow, get_ks()):
            yield line
        return

    def get_line_iter_2(self):
        # Newer variant

        mod = self.mod

        srcrow = self.srcrow

        indinc = 1
        if srcrow.depth >= 10:
            indinc = 1

        lno = [0]

        seen = {}

        indir = 1

        if indir == 1:
            max_ixlstr_len = 0
            max_str_len_set = 0
            row = srcrow
            while row:
                if len(row.ixlstr) > max_ixlstr_len:
                    max_ixlstr_len = len(row.ixlstr)
                if len(str(len(row.set.nodes))) > max_str_len_set:
                    max_str_len_set = len(str(len(row.set.nodes)))
                row = row.parent

        def genlines(row, part, idx):

            seen[part.nodes, row.depth] = lno[0]
            sidx = row.ixlstr
            idxs = '[%d]' % idx
            if indir < 0:
                indent = (row.depth)*indinc
                sidx = '%s%s%s' % (
                    sidx, ' '*(6+indent-len(sidx)-len(idxs)), idxs)
                if row.parent is None:
                    sidx += ' == %s' % part.brief

            else:
                #idxs = ('[%.'+str(max_str_len_set)+'d]')%idx
                sidx = '%s%s%s' % (sidx,
                                   ' '*(3+max_str_len_set +
                                        max_ixlstr_len-len(sidx)-len(idxs)),
                                   idxs)
                sidx += ' ' * (srcrow.depth + 1 - row.depth)
                if row.parent is not None:
                    sidx += '@'
                else:
                    sidx += '= %s' % part.brief

            if row.parent is None:
                #vline += ' == %s'%self.get_str_of_path_component_singleton(part)
                vline = '%2s: %s' % (lno[0], sidx)
                lno[0] += 1
                yield ('STOP_AFTER', vline)
                return

            referents = part.referents & row.parent.set & self.andsetbyname[row.parent.ixlstr]
            relations = mod.Path.relations
            iso = mod.Use.iso
            s = part.theone
            t = [(relations(s, p.theone), p.by(referents.er), i)
                 for (i, p) in enumerate(referents.byid.parts)]
            for (rels, p, i) in t:
                relstrings = []
                for rel in rels:
                    if rel is mod.Path.identity:
                        continue
                    if rel is mod.Path.norelation:
                        k = '??'
                    else:
                        k = str(rel) % ''
                    relstrings.append(k)

                relsstr = ' / '.join(relstrings)
                seenlno = seen.get((p.nodes, row.parent.depth))
                vline = '%2s: %s' % (lno[0], sidx)
                lno[0] += 1
                if seenlno is not None:
                    relsstr += ' -> #%d' % seenlno
                    yield ('STOP_AFTER', vline + ' ' + relsstr)
                else:
                    yield vline + ' ' + relsstr
                    for line in genlines(row.parent, p, i):
                        yield line

        for i, p in enumerate((srcrow.set & self.andsetbyname[srcrow.ixlstr]).byid.parts):
            for line in genlines(srcrow, p, i):
                yield line


class RefPatIter:
    def __init__(self, rp, start=0):
        self.rp = rp
        self._hiding_tag_ = rp._hiding_tag_
        self.ix = start

    def __iter__(self):
        return self

    def __next__(self):
        try:
            x = self.rp[self.ix]
        except IndexError:
            raise StopIteration
        self.ix += 1
        return x


class RefPatRow:
    def __init__(self, rp, kindset, seenline, ixl, parent):
        self.rp = rp
        self._hiding_tag_ = rp._hiding_tag_
        self.kindset = kindset
        self.kind, self.set = kindset
        assert self.set <= self.kind
        self.seenline = seenline
        self.ixl = ixl[:]
        self.parent = parent
        if parent is not None:
            self.depth = parent.depth + 1
        else:
            self.depth = 0

        self.index = 0
        self.maxdepth = rp.depth
        self.max_str_len = rp.mod.line_length
        self.ixlstr = ixl_as_str(ixl)
        self.isready = 0
        self.children = []

    def __str__(self):
        prestr = '%2d: %s ' % (self.index, self.ixlstr)

        if self.index & 1:
            fillpat = ' ' * 100
        else:
            fillpat = '-'*100

        lps = len(prestr)
        fill = fillpat[lps:9+self.depth]

        if self.seenline:
            ref = '[^ %s]' % self.seenline.index
        elif self.isroot:
            ref = '[R]'
        elif self.depth > 0 and self.set <= self.rp.stopkind:
            ref = '[S]'
        elif self.depth < self.maxdepth:
            ref = '[-]'
        else:
            ref = '[+]'

        prefix = '%s%s %s ' % (prestr, fill, ref)
        return '%s%s' % (prefix, self.getsummary(self.max_str_len-len(prefix)))

    def getchild(self, ix):
        while ix >= len(self.children) and not self.isready:
            self.rp.generate(len(self.rp.lines))
        return self.children[ix]

    def getsummary(self, max_len):
        kind, set = self.kind, self.set
        summary = set.fam.get_str_refpat(set, kind, max_len)
        return summary


class ReferencePattern:
    __doc__ = '<Help Text'
    help = """\
Methods




"""

    def __init__(self, mod, set, depth, er, relimg, bf, stopkind, nocyc):
        self.mod = mod
        self._hiding_tag_ = mod._hiding_tag_
        self.View = mod.View
        self.set = set
        self.depth = depth
        self.er = er
        self.bf = bf
        self.stopkind = stopkind
        self.nocyc = nocyc
        self.is_initialized = 0

        self.totcount = set.count
        self.totsize = set.indisize
        self.kind = set.kind
        self.kindset = (self.kind, self.set)
        self.relimg = relimg
        self.top = self

        mod.OutputHandling.setup_printing(self)
        self.View.referrers_add_target(self)
        self.reset_nogc()
        self.is_initialized = 1

    def __getattr__(self, s):
        if not self.is_initialized:
            raise AttributeError(s)

        try:
            return getattr(self.__class__, s)
        except AttributeError:
            pass
        try:
            row = self.get_row_named(s)
        except ValueError:
            raise AttributeError(s)
        return row.set

    def __getitem__(self, ix):
        return self.get_row_indexed(ix).set

    def __iter__(self, start=0):
        return RefPatIter(self, start)

    def __len__(self):
        self.generate()
        return len(self.lines)

    def _oh_get_label(self):
        return 'Reference Pattern by <' + self.er.classifier.get_byname() + '>.'

    def _oh_get_more_state_msg(self, startindex, lastindex):
        if self.isfullygenerated:
            msg = '%d more lines. ' % (len(self.lines) - 1-lastindex,)
        else:
            msg = ''
        return msg

    def _oh_get_line_iter(self):
        it = self.iterlines(0)
        for el in it:
            yield str(el)

    def generate(self, ix=None):
        while ix is None or ix < 0 or ix >= len(self.lines):
            try:
                self.lines.append(next(self.lg))
            except StopIteration:
                self.isfullygenerated = 1
                return
            self.lines[-1].index = len(self.lines) - 1

    def get_row(self, key):
        try:
            [][key]
        except TypeError:
            return self.get_row_named(key)
        except IndexError:
            return self.get_row_indexed(key)

    def get_row_indexed(self, ix):
        self.generate(ix)
        return self.lines[ix]

    def get_row_named(self, name):
        row = self.get_row_indexed(0)
        for ix in str_as_ixl(name):
            try:
                row = row.getchild(ix)
            except IndexError:
                raise ValueError(
                    'Reference pattern has no row named %r' % name)
        return row

    def iterlines(self, start=None):
        if start is None:
            start = 0
        while 1:
            try:
                yield self.get_row_indexed(start)
            except IndexError:
                return
            start += 1

    def linegenerator(self, xxx_todo_changeme, ixl, parent=None):
        (kind, set) = xxx_todo_changeme
        seenline = self.seensets.get(set.nodes)
        ixl = list(ixl)
        line = RefPatRow(self, (kind, set), seenline=seenline,
                         ixl=ixl, parent=parent)
        children = self.get_children(line)
        line.isroot = not children
        if seenline is None:
            self.seensets[set.nodes] = line
        if parent is not None:
            parent.children.append(line)
        yield line

        depth = line.depth
        if (not seenline and depth < self.depth and
                (depth == 0 or not (set <= self.stopkind))):
            for i, cs in enumerate(children):
                ixl.append(i)
                for rl in self.linegenerator(cs, ixl, line):
                    yield rl
                ixl.pop()
        line.isready = 1

    def get_children(self, line):
        (kind, set) = line.kindset
        chset = self.relimg(set)
        if self.nocyc:
            while line is not None:
                chset -= line.set
                line = line.parent
        return [(row.kind, row.set)
                for row in self.get_partition(chset, self.er).get_rows()]

    def get_partition(self, set, er):
        p = self.mod.Part.partition(set, er)
        return p

    def paths(self, key, **kwds):
        return Paths(self.mod, self, key, **kwds)

    def reset(self):
        self.reset_nogc()
        self.printer.reset()
        self.mod._root.gc.collect()

    def reset_nogc(self):
        self.isfullygenerated = 0
        self.seensets = {}
        self.lines = []
        self.lg = self.linegenerator(self.kindset, [])
        self.lastindex = None


class _GLUECLAMP_:
    _preload_ = ('_hiding_tag_',)

    depth = 7
    line_length = 80

    _uniset_exports = ('rp',)
    _chgable_ = ('er',
                 'depth',
                 'line_length',
                 )

    # 'module imports'

    _imports_ = (
        '_parent:OutputHandling',
        '_parent:Part',
        '_parent:Path',
        '_parent:UniSet',
        '_parent:Use',
        '_parent:View',
        '_parent.View:_hiding_tag_',
    )

    #

    def _get_er(self): return self.Use.Clodo

    def _get_stopkind(self):
        hp = self.Use
        return (
            hp.Type.Module |
            hp.Type.Type |
            hp.Type.Module.dictof |
            hp.Type.Type.dictof |
            hp.Type.Code |
            hp.Type.Frame
        )

    def rp(self, X, depth=None, er=None, imdom=0, bf=0, src=None, stopkind=None,
           nocyc=False, ref=None):
        """rp(X, depth=None, er=None, imdom=0, bf=0, src=None, stopkind=None, nocyc=False, ref=None)
Reference pattern forming.
Arguments
        X       Set of objects for which a reference pattern is sought.
        depth   The depth to which the pattern will be generated. The
                default is taken from depth of this module.
        er      The equivalence relation to partition the referrers. The default
                is Clodo.
        imdom   If true, the immediate dominators will be used instead
                of the referrers. This will take longer time to calculate,
                but may be useful to reduce the complexity of the reference
                pattern.
        bf      If true, the pattern will be printed in breadth-first
                order instead of depth-first. (Experimental.)
        src     If specified, an alternative reference source instead
                of the default root.
        stopkind
        nocyc
        ref

Description
        Return a reference pattern object based on the objects in the set X.
        The reference pattern object is of class ReferencePattern. It is
        described in XXX.
"""
        if depth is None:
            depth = self.depth
        X = self.UniSet.idset_adapt(X)
        if src is not None:
            # Creates an entire new guppy tree
            # Mostly for testing purposes -
            # can likely cause type problems generally
            src = self.UniSet.idset_adapt(src)
            self = self._root.guppy.Root().guppy.heapy.RefPat
            self.View.root = tuple(src.nodes)
            X = self.Use.idset(X.nodes)

        if er is None:
            er = self.er  # NEEDS to be loaded after new Classifier created
        if imdom:
            def relimg(X): return X.imdom
        elif ref is not None:
            if ref == 'gc':
                def relimg(X): return X.referrers_gc
            elif ref == 'gcx':
                relimg = (lambda x: x.referrers_gc
                          - self._root.guppy.sets.ImmNodeSet
                          - self._parent.heapyc.NodeGraph
                          - self.View.ObservationList)
            elif ref == 'imdom':
                def relimg(X): return X.imdom
            elif callable(ref):
                relimg = ref
            else:
                raise ValueError(
                    "ref should be 'gc', 'gcx', 'imdom', or a callable")
        else:
            def relimg(X): return X.referrers
        if stopkind is None:
            stopkind = self.stopkind
        rp = ReferencePattern(self, X, depth, er, relimg, bf, stopkind, nocyc)
        return rp



class Graph:
    def __init__(self, objects, arrows):
        self.objects = objects  # Sequence of objects
        self.arrows = arrows    # Map[name] ->pair(object, object)

    def source(self, x):
        return self.arrows[x][0]

    def target(self, x):
        return self.arrows[x][1]

    def get_dual(self):
        objects = self.objects
        arrows = dict([(arrow, (tgt, src))
                       for (arrow, (src, tgt)) in list(self.arrows.items())])
        return self.__class__(objects, arrows)


class Cat:
    # Category presented by a graph (with objects and generators) and relations.
    def __init__(self, graph, relations):
        # category is defined by the parameters:
        #    graph.objects: sequenceof(O)
        #    graph.arrows: dict mapping(A, pairof(O in objects))
        #    relations: sequence(pairof(sequence(A), sequence(A)))
        self.graph = graph
        self.relations = relations

    def get_dual(self):
        graph = self.graph.get_dual()
        relations = dual_relations(self.relations)
        return self.__class__(graph, relations)


class Functor:
    def __init__(self, fo, fa, src=None, tgt=None):
        self.fo = adapt_function(fo)
        self.fa = adapt_function(fa)
        self.src = src
        self.tgt = tgt


class Function:
    def __init__(self, map, src, tgt):
        f = getattr(map, '__getitem__', None)
        if callable(f):
            pass
        else:
            f = map
            if not callable(f):
                raise TypeError(
                    'Function: map is neither callable or indexable')
        self.f = f
        self.src = src
        self.tgt = tgt

    def __getitem__(self, *args):
        return self.f(*args)

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def __str__(self):
        return '%s(%s, %s, %s)' % (self.__class__, self.src, self.tgt, self.f)

    def asdict(self):
        return dict([(x, self[x]) for x in self.src])

    def items(self):
        return [(x, self[x]) for x in self.src]

    def keys(self):
        return list(self.src)

    def values(self):
        return [v for (k, v) in list(self.items())]


class Identity(Function):
    def __init__(self, src):
        Function.__init__(lambda x: x, src, src)


def check_graph(G):
    # Check that G is a valid graph object
    # with arrows that have all source and target in G.objects

    Gob = G.objects
    for a in G.arrows:
        if not G.source(a) in Gob:
            raise ValueError(
                'Arrow %r has source %r not in graph objects' % (a, G.source(a)))
        if not G.target(a) in Gob:
            raise ValueError(
                'Arrow %r has target %r not in graph objects' % (a, G.target(a)))


def check_rules(R, G):
    # Check that the rules in R contain valid composing arrows in graph G

    coms = []
    for (left, right) in R:
        coms.append(left)
        coms.append(right)

    for com in coms:
        a0 = None
        for a in com:
            if a not in G.arrows:
                raise ValueError(
                    'Arrow %r, used in a rule, is not a valid arrow' % (a,))
            if a0 is not None:
                if G.source(a) != G.target(a0):
                    raise ValueError('''\
Source of arrow %r (%r) does not match target of arrow %r (%r)''' % (
                        a, G.source(a), a0, G.target(a0)))
            a0 = a


def check_cat(C):
    check_graph(C.graph)
    check_rules(C.relations, C.graph)


def oarcat(objects, arrows, relations):
    return Cat(Graph(objects, arrows), relations)


def adapt_function(f):
    if not isinstance(f, Function):
        if isinstance(f, dict):
            src = list(f.keys())
            tgt = list(f.values())
        else:
            src = None
            tgt = None
        f = Function(f, src, tgt)
    return f


def dual_relations(relations):
    dual = []
    for (a, b) in relations:
        a = list(a)
        b = list(b)
        a.reverse()
        b.reverse()
        dual.append((tuple(a), tuple(b)))
    return dual



from guppy.sets.setsc import BitSet        # base bitset type
from guppy.sets.setsc import ImmBitSet     # immutable bitset type
from guppy.sets.setsc import immbit        # immutable bitset singleton constructor
from guppy.sets.setsc import immbitrange   # immutable bitset range constructor
from guppy.sets.setsc import immbitset     # immutable bitset constructor
from guppy.sets.setsc import MutBitSet     # mutable bitset
from guppy.sets.setsc import NodeSet       # base nodeset type
from guppy.sets.setsc import ImmNodeSet    # immmutable nodeset type
from guppy.sets.setsc import MutNodeSet    # mutable nodeset type

from .setsc import _bs
_bs.__module__ = 'guppy.sets'  # ..to be able to set it.


# Define some constructors.
# Constructor names are lower case.
# Some constructors are equal to types.
# But this connection depends on the implementation.
# So one may wish the user to not depend on this.

mutbitset = MutBitSet
immnodeset = ImmNodeSet
mutnodeset = MutNodeSet


def mutnodeset_union(iterable):
    "Return a mutable nodeset which is the union of all nodesets in iterable."
    set = mutnodeset()
    for it in iterable:
        set |= it
    return set


def immnodeset_union(iterable, *args):
    "Return an immmutable nodeset which is the union of all nodesets in iterable."
    set = mutnodeset_union(iterable)
    return immnodeset(set, *args)


# Make attributes assignable by reading one;
# this is getting around a bug in Python 2.3.3
# and should be harmless in any version.


try:
    mutnodeset()._hiding_tag_
except AttributeError:
    pass



class GsmlHandler:
    # To be mixed in with something like HTMLParser.HTMLParser

    def handle_starttag(self, tag, attrs):
        self.stack.append(self.out)
        self.out = []
        if attrs:
            at = []
            for k, v in attrs:
                at.append(self.mod.node_of_taci(k, v))
            self.out.append(self.mod.node_of_taci('with', '', at))

    def handle_endtag(self, tag):
        node = self.mod.node_of_taci(tag, '', self.out)
        self.out = self.stack.pop()
        self.out.append(node)

    def handle_charref(self, name):
        if name[:1] == "x":
            char = int(name[1:], 16)
            name = '0'+name
        else:
            char = int(name)
        if 0 <= char < 128:
            char = chr(char)
            self.handle_data(char)
        else:
            self.out.append(self.mod.node_of_taci('char', name))

    def handle_entityref(self, name):
        if name not in self.mod.entitydefs:
            self.unknown_entityref(name)
        self.out.append(self.mod.node_of_taci('char', name))

    def unknown_entityref(self, name):
        raise SyntaxError('Unknown entity ref: %r' % name)

    def handle_data(self, data):
        # data = data.strip()
        if data.strip():
            self.out.extend(self.mod.nodes_of_text(data))

    def handle_comment(self, data):
        self.out.append(self.mod.node_of_taci('comment', data, (), 0))

    def handle_decl(self, decl):
        self.out.append(self.mod.node_of_taci('html_declaration', decl))

    def handle_pi(self, data):
        self.out.append(self.mod.node_of_taci('processing_instruction', data))


class _GLUECLAMP_:
    _imports_ = (
        '_root.html.parser:HTMLParser',
        '_parent.SpecNodes:node_of_taci',
        '_parent.SpecNodes:nodes_of_text',
        '_root.htmlentitydefs:entitydefs',
    )

    encoding = "iso-8859-1"

    def node_of_gsml(self, text):
        class Parser(GsmlHandler, self.HTMLParser):
            def __init__(self, mod):
                mod.HTMLParser.__init__(self)
                self.mod = mod
                self.out = []
                self.stack = []

        p = Parser(self)
        p.feed(text)
        p.close()
        if p.stack:
            raise SyntaxError('Missing end tag')
        node = self.node_of_taci('block', '', p.out, 0)
        return node

    def _test_main_(self):
        x = """
<!DOCTYPE html...>
This is an <em> emphasized </em> word.
See also <a href="guppy-pe.sf.net"> Guppy. </a>
Defined as <use_my_macro/>.

Handle char ref: &#100;.
Handle char ref: &lt;.

<!-- A comment -->

<?A processing instruction>

<namespace:tag />
        """

        node = self.node_of_gsml(x)
        # FIXME: Assert equals to something?
        print(node)
